import datetime
from collections.abc import AsyncIterable, AsyncIterator, Iterable, Iterator, Mapping
from http.cookies import SimpleCookie
from io import BytesIO
from json import JSONEncoder
from typing import Any, Literal, NoReturn, TypeVar, overload, type_check_only

from django.utils.datastructures import CaseInsensitiveMapping, _PropertyDescriptor
from django.utils.functional import _StrOrPromise, cached_property
from typing_extensions import override

class BadHeaderError(ValueError): ...

_Z = TypeVar("_Z")

class ResponseHeaders(CaseInsensitiveMapping[str]):
    def __init__(self, data: Mapping[str, str]) -> None: ...
    def _convert_to_charset(self, value: bytes | str | int, charset: str, mime_encode: bool = ...) -> str: ...
    def __delitem__(self, key: str) -> None: ...
    def __setitem__(self, key: str, value: str | bytes | int) -> None: ...
    def pop(self, key: str, default: _Z = ...) -> _Z | tuple[str, str]: ...
    def setdefault(self, key: str, value: str | bytes | int) -> None: ...

class HttpResponseBase:
    status_code: int
    streaming: bool
    cookies: SimpleCookie
    closed: bool
    headers: ResponseHeaders
    def __init__(
        self,
        content_type: str | None = ...,
        status: int | None = ...,
        reason: str | None = ...,
        charset: str | None = ...,
        headers: Mapping[str, str] | None = ...,
    ) -> None: ...
    @property
    def reason_phrase(self) -> str: ...
    @reason_phrase.setter
    def reason_phrase(self, value: str) -> None: ...
    @property
    def charset(self) -> str: ...
    @charset.setter
    def charset(self, value: str) -> None: ...
    def serialize_headers(self) -> bytes: ...
    __bytes__ = serialize_headers
    def __setitem__(self, header: str, value: str | bytes | int) -> None: ...
    def __delitem__(self, header: str) -> None: ...
    def __getitem__(self, header: str) -> str: ...
    def has_header(self, header: str) -> bool: ...
    def __contains__(self, header: str) -> bool: ...
    def items(self) -> Iterable[tuple[str, str]]: ...
    @overload
    def get(self, header: str, alternate: str) -> str: ...
    @overload
    def get(self, header: str, alternate: None = None) -> str | None: ...
    def set_cookie(
        self,
        key: str,
        value: str = ...,
        max_age: int | datetime.timedelta | None = ...,
        expires: str | datetime.datetime | None = ...,
        path: str = ...,
        domain: str | None = ...,
        secure: bool = ...,
        httponly: bool = ...,
        samesite: Literal["Lax", "Strict", "None", False] | None = ...,
    ) -> None: ...
    def setdefault(self, key: str, value: str) -> None: ...
    def set_signed_cookie(self, key: str, value: str, salt: str = ..., **kwargs: Any) -> None: ...
    def delete_cookie(
        self,
        key: str,
        path: str = ...,
        domain: str | None = ...,
        samesite: Literal["Lax", "Strict", "None", False] | None = ...,
    ) -> None: ...
    def make_bytes(self, value: object) -> bytes: ...
    def close(self) -> None: ...
    def write(self, content: str | bytes) -> None: ...
    def flush(self) -> None: ...
    def tell(self) -> int: ...
    def readable(self) -> bool: ...
    def seekable(self) -> bool: ...
    def writable(self) -> bool: ...
    def writelines(self, lines: Iterable[object]) -> None: ...

    # Fake methods that are implemented by all subclasses
    @type_check_only
    def __iter__(self) -> Iterator[bytes]: ...
    @type_check_only
    def getvalue(self) -> bytes: ...

class HttpResponse(HttpResponseBase, Iterable[bytes]):
    content = _PropertyDescriptor[object, bytes]()
    csrf_cookie_set: bool
    sameorigin: bool
    test_server_port: str
    test_was_secure_request: bool
    xframe_options_exempt: bool
    def __init__(self, content: object = ..., *args: Any, **kwargs: Any) -> None: ...
    def serialize(self) -> bytes: ...
    __bytes__ = serialize
    @override
    def __iter__(self) -> Iterator[bytes]: ...
    @override
    def getvalue(self) -> bytes: ...
    @cached_property
    def text(self) -> str: ...

class StreamingHttpResponse(HttpResponseBase, Iterable[bytes], AsyncIterable[bytes]):
    is_async: bool
    streaming_content = _PropertyDescriptor[
        Iterable[object] | AsyncIterable[object], Iterator[bytes] | AsyncIterator[bytes]
    ]()
    def __init__(
        self, streaming_content: Iterable[object] | AsyncIterable[object] = ..., *args: Any, **kwargs: Any
    ) -> None: ...
    @override
    def __iter__(self) -> Iterator[bytes]: ...
    @override
    def __aiter__(self) -> AsyncIterator[bytes]: ...
    @override
    def getvalue(self) -> bytes: ...
    @property
    def content(self) -> NoReturn: ...
    @property
    def text(self) -> NoReturn: ...

class FileResponse(StreamingHttpResponse):
    file_to_stream: BytesIO | None
    block_size: int
    as_attachment: bool
    filename: str
    def __init__(self, *args: Any, as_attachment: bool = ..., filename: str = ..., **kwargs: Any) -> None: ...
    def set_headers(self, filelike: BytesIO) -> None: ...

class HttpResponseRedirectBase(HttpResponse):
    allowed_schemes: list[str]
    def __init__(
        self, redirect_to: _StrOrPromise, preserve_request: bool = False, *args: Any, **kwargs: Any
    ) -> None: ...
    @property
    def url(self) -> str: ...

class HttpResponseRedirect(HttpResponseRedirectBase):
    status_code_preserve_request: int

class HttpResponsePermanentRedirect(HttpResponseRedirectBase):
    status_code_preserve_request: int

class HttpResponseNotModified(HttpResponse):
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...

class HttpResponseBadRequest(HttpResponse): ...
class HttpResponseNotFound(HttpResponse): ...
class HttpResponseForbidden(HttpResponse): ...

class HttpResponseNotAllowed(HttpResponse):
    def __init__(self, permitted_methods: Iterable[str], *args: Any, **kwargs: Any) -> None: ...

class HttpResponseGone(HttpResponse): ...
class HttpResponseServerError(HttpResponse): ...
class Http404(Exception): ...

class JsonResponse(HttpResponse):
    def __init__(
        self,
        data: Any,
        encoder: type[JSONEncoder] = ...,
        safe: bool = ...,
        json_dumps_params: dict[str, Any] | None = ...,
        **kwargs: Any,
    ) -> None: ...



import sys
from ctypes import Structure, Union
from typing import Any

LOCK_SH: int
LOCK_NB: int
LOCK_EX: int

if sys.platform == "win32":
    ULONG_PTR: Any
    PVOID: Any
    class _OFFSET(Structure): ...
    class _OFFSET_UNION(Union): ...
    class OVERLAPPED(Structure): ...

def lock(f: Any, flags: int) -> bool: ...
def unlock(f: Any) -> bool: ...

__all__ = ("LOCK_EX", "LOCK_NB", "LOCK_SH", "lock", "unlock")



from typing import Any

from django.http.request import HttpRequest
from django.http.response import FileResponse, HttpResponse

def serve(request: HttpRequest, path: str, insecure: bool = False, **kwargs: Any) -> HttpResponse | FileResponse: ...



from functools import lru_cache
from typing import Any

from django.db.backends.postgresql.base import DatabaseWrapper as Psycopg2DatabaseWrapper
from psycopg import BaseConnection
from psycopg.adapt import Dumper
from psycopg.pq import Format
from typing_extensions import override

class BaseBinaryDumper(Dumper):
    format: Format
    @override
    def dump(self, obj: Any) -> bytes: ...

class BaseTextDumper(Dumper):
    @override
    def dump(self, obj: Any) -> bytes: ...

class DatabaseWrapper(Psycopg2DatabaseWrapper):
    SchemaEditorClass: Any
    features: Any
    ops: Any
    introspection: Any
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    @override
    def prepare_database(self) -> None: ...
    def register_geometry_adapters(self, pg_connection: BaseConnection[bytes], clear_caches: bool = False) -> None: ...

class GeometryType: ...
class GeographyType: ...
class RasterType: ...

@lru_cache
def postgis_adapters(
    geo_oid: int | None,
    geog_oid: int | None,
    raster_oid: int | None,
) -> tuple[type[Dumper], type[Dumper]]: ...



from collections.abc import Awaitable, Callable
from typing import Any, ClassVar, Literal

from django.contrib.auth.models import _AnyUser
from django.http import HttpRequest, HttpResponseBase, HttpResponseRedirect
from django.utils.deprecation import MiddlewareMixin, _AsyncGetResponseCallable, _GetResponseCallable

def get_user(request: HttpRequest) -> _AnyUser: ...
async def auser(request: HttpRequest) -> _AnyUser: ...

class AuthenticationMiddleware(MiddlewareMixin):
    def process_request(self, request: HttpRequest) -> None: ...

class LoginRequiredMiddleware(MiddlewareMixin):
    redirect_field_name: ClassVar[str]

    def process_view(
        self,
        request: HttpRequest,
        view_func: Callable[..., HttpResponseBase],
        view_args: tuple[Any, ...],
        view_kwargs: dict[Any, Any],
    ) -> HttpResponseBase | None: ...
    def get_login_url(self, view_func: Callable[..., HttpResponseBase]) -> str: ...
    def get_redirect_field_name(self, view_func: Callable[..., HttpResponseBase]) -> str: ...
    def handle_no_permission(
        self, request: HttpRequest, view_func: Callable[..., HttpResponseBase]
    ) -> HttpResponseRedirect: ...

class RemoteUserMiddleware:
    header: ClassVar[str]
    force_logout_if_no_header: ClassVar[bool]
    sync_capable: ClassVar[Literal[True]]
    async_capable: ClassVar[Literal[True]]
    get_response: _GetResponseCallable | _AsyncGetResponseCallable
    is_async: bool
    def __init__(self, get_response: _GetResponseCallable | _AsyncGetResponseCallable) -> None: ...
    def __call__(self, request: HttpRequest) -> HttpResponseBase | Awaitable[HttpResponseBase]: ...
    async def __acall__(self, request: HttpRequest) -> HttpResponseBase: ...
    def clean_username(self, username: str, request: HttpRequest) -> str: ...
    def process_request(self, request: HttpRequest) -> None: ...
    async def aprocess_request(self, request: HttpRequest) -> None: ...

class PersistentRemoteUserMiddleware(RemoteUserMiddleware):
    force_logout_if_no_header: ClassVar[bool]



from typing import Any

from django.contrib.admin.checks import InlineModelAdminChecks
from django.contrib.admin.options import InlineModelAdmin
from django.contrib.contenttypes.forms import BaseGenericInlineFormSet
from django.db.models.base import Model

class GenericInlineModelAdminChecks(InlineModelAdminChecks):
    def _check_exclude_of_parent_model(self, obj: GenericInlineModelAdmin, parent_model: type[Model]) -> list[Any]: ...
    def _check_relation(self, obj: GenericInlineModelAdmin, parent_model: type[Model]) -> list[Any]: ...

class GenericInlineModelAdmin(InlineModelAdmin):
    ct_field: str
    ct_fk_field: str
    formset: type[BaseGenericInlineFormSet]  # type: ignore[assignment]

class GenericStackedInline(GenericInlineModelAdmin):
    template: str

class GenericTabularInline(GenericInlineModelAdmin):
    template: str



from typing import Any

from django.apps.config import AppConfig
from django.apps.registry import Apps
from django.db import migrations
from django.db.backends.sqlite3.schema import DatabaseSchemaEditor
from django.db.migrations.migration import Migration
from django.db.migrations.state import StateApps

class RenameContentType(migrations.RunPython):
    app_label: str
    old_model: str
    new_model: str
    def __init__(self, app_label: str, old_model: str, new_model: str) -> None: ...
    def rename_forward(self, apps: StateApps, schema_editor: DatabaseSchemaEditor) -> None: ...
    def rename_backward(self, apps: StateApps, schema_editor: DatabaseSchemaEditor) -> None: ...

def inject_rename_contenttypes_operations(
    plan: list[tuple[Migration, bool]] | None = None, apps: Apps = ..., using: str = "default", **kwargs: Any
) -> None: ...
def create_contenttypes(
    app_config: AppConfig,
    verbosity: int = 2,
    interactive: bool = True,
    using: str = "default",
    apps: Apps = ...,
    **kwargs: Any,
) -> None: ...



from typing import Any

from django.contrib.gis.geos.libgeos import GEOSFuncFactory

class DblFromGeom(GEOSFuncFactory):
    restype: Any
    errcheck: Any

geos_area: Any
geos_distance: Any
geos_length: Any
geos_isvalidreason: Any

__all__ = ["geos_area", "geos_distance", "geos_isvalidreason", "geos_length"]



from mypy.plugin import MethodContext
from mypy.types import AnyType, Instance, ProperType, TypeOfAny, get_proper_type
from mypy.types import Type as MypyType

from mypy_django_plugin.django.context import DjangoContext
from mypy_django_plugin.exceptions import UnregisteredModelError
from mypy_django_plugin.lib import fullnames, helpers


def typecheck_queryset_filter(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    django_model = helpers.get_model_info_from_qs_ctx(ctx, django_context)
    if django_model is None:
        return ctx.default_return_type

    # Expected formal arguments for filter methods are `*args` and `**kwargs`. We'll only typecheck
    # `**kwargs`, which means that `arg_names[1]` is what we're interested in.
    lookup_kwargs = ctx.arg_names[1] if len(ctx.arg_names) >= 2 else []
    provided_lookup_types = ctx.arg_types[1] if len(ctx.arg_types) >= 2 else []

    for lookup_kwarg, provided_type in zip(lookup_kwargs, provided_lookup_types, strict=False):
        if lookup_kwarg is None:
            continue
        provided_type = get_proper_type(provided_type)
        if isinstance(provided_type, Instance) and provided_type.type.has_base(
            fullnames.COMBINABLE_EXPRESSION_FULLNAME
        ):
            provided_type = resolve_combinable_type(provided_type, django_context)

        lookup_type: MypyType
        try:
            lookup_type = django_context.resolve_lookup_expected_type(
                ctx, django_model.cls, lookup_kwarg, django_model.typ
            )
        except UnregisteredModelError:
            lookup_type = AnyType(TypeOfAny.from_error)
        # Managers as provided_type is not supported yet
        if isinstance(provided_type, Instance) and helpers.has_any_of_bases(
            provided_type.type, (fullnames.MANAGER_CLASS_FULLNAME, fullnames.QUERYSET_CLASS_FULLNAME)
        ):
            return ctx.default_return_type

        helpers.check_types_compatible(
            ctx,
            expected_type=lookup_type,
            actual_type=provided_type,
            error_message=f"Incompatible type for lookup {lookup_kwarg!r}:",
        )

    return ctx.default_return_type


def resolve_combinable_type(combinable_type: Instance, django_context: DjangoContext) -> ProperType:
    if combinable_type.type.fullname != fullnames.F_EXPRESSION_FULLNAME:
        # Combinables aside from F expressions are unsupported
        return AnyType(TypeOfAny.explicit)

    return django_context.resolve_f_expression_type(combinable_type)



from typing import Any, NamedTuple

from django.db.backends.base.introspection import BaseDatabaseIntrospection
from django.db.backends.oracle.base import DatabaseWrapper

FieldInfo: Any

class TableInfo(NamedTuple):
    name: str
    type: str
    comment: str | None

class DatabaseIntrospection(BaseDatabaseIntrospection):
    connection: DatabaseWrapper
    cache_bust_counter: int
    data_types_reverse: dict[int, str]



import decimal
import uuid
from collections.abc import Callable, Iterable, Mapping, Sequence
from datetime import date, time, timedelta
from datetime import datetime as real_datetime
from typing import Any, ClassVar, Generic, Literal, Protocol, TypeAlias, TypeVar, overload, type_check_only

from django import forms
from django.core import validators  # due to weird mypy.stubtest error
from django.core.checks import CheckMessage
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models import Model
from django.db.models.expressions import Col, Combinable, Expression, Func
from django.db.models.fields.reverse_related import ForeignObjectRel
from django.db.models.query import _OrderByFieldName
from django.db.models.query_utils import Q, RegisterLookupMixin
from django.db.models.sql.compiler import SQLCompiler, _AsSqlType, _ParamsT
from django.utils.choices import BlankChoiceIterator, _Choice, _ChoiceNamedGroup, _ChoicesCallable, _ChoicesInput
from django.utils.datastructures import DictWrapper
from django.utils.functional import _Getter, _StrOrPromise, cached_property
from typing_extensions import Self, override

class Empty: ...
class NOT_PROVIDED: ...

BLANK_CHOICE_DASH: list[tuple[str, str]]

_ChoicesList: TypeAlias = Sequence[_Choice] | Sequence[_ChoiceNamedGroup]
_LimitChoicesTo: TypeAlias = Q | dict[str, Any]
_LimitChoicesToCallable: TypeAlias = Callable[[], _LimitChoicesTo]

_F = TypeVar("_F", bound=Field, covariant=True)

@type_check_only
class _FieldDescriptor(Protocol[_F]):
    """
    Accessing fields of a model class (not instance) returns an object conforming to this protocol.
    Depending on field type this could be DeferredAttribute, ForwardManyToOneDescriptor, FileDescriptor, etc.
    """

    @property
    def field(self) -> _F: ...

_AllLimitChoicesTo: TypeAlias = _LimitChoicesTo | _LimitChoicesToCallable | _ChoicesCallable  # noqa: PYI047
_ErrorMessagesMapping: TypeAlias = Mapping[str, _StrOrPromise]
_ErrorMessagesDict: TypeAlias = dict[str, _StrOrPromise]

# __set__ value type
_ST = TypeVar("_ST", contravariant=True)
# __get__ return type
_GT = TypeVar("_GT", covariant=True)

class Field(RegisterLookupMixin, Generic[_ST, _GT]):
    """
    Typing model fields.

    How does this work?
    Let's take a look at the self-contained example
    (it is way easier than our django implementation, but has the same concept).

    To understand this example you need:
    1. Be familiar with descriptors: https://docs.python.org/3/howto/descriptor.html
    2. Follow our explanation below

    Let's start with defining our fake model class and fake integer field.

    .. code:: python

        from typing import Generic, Union

        class Model(object):
            ...

        _SetType = Union[int, float]  # You can assign ints and floats
        _GetType = int  # access type is always `int`

        class IntField(object):
            def __get__(self, instance: Model, owner) -> _GetType:
                ...

            def __set__(self, instance, value: _SetType) -> None:
                ...

    Now, let's create our own example model,
    this would be something like ``User`` in your own apps:

    .. code:: python

        class Example(Model):
            count = IntField()

    And now, lets test that our reveal type works:

    .. code:: python

        example = Example()
        reveal_type(example.count)
        # Revealed type is "builtins.int"

        example.count = 1.5  # ok
        example.count = 'a'
        # Incompatible types in assignment
        # (expression has type "str", variable has type "Union[int, float]")

    Notice, that this is not magic. This is how descriptors work with ``mypy``.

    We also need ``_pyi_private_set_type`` attributes
    and friends to help inside our plugin.
    It is required to enhance parts like ``filter`` queries.
    """

    _pyi_private_set_type: Any
    _pyi_private_get_type: Any
    _pyi_lookup_exact_type: Any

    help_text: _StrOrPromise
    attname: str
    auto_created: bool
    primary_key: bool
    remote_field: ForeignObjectRel | None
    is_relation: bool
    related_model: type[Model] | Literal["self"] | None
    generated: ClassVar[bool]
    one_to_many: bool | None
    one_to_one: bool | None
    many_to_many: bool | None
    many_to_one: bool | None
    max_length: int | None
    model: type[Model]
    name: str
    verbose_name: _StrOrPromise
    description: _StrOrPromise | _Getter[_StrOrPromise]
    blank: bool
    null: bool
    editable: bool
    empty_strings_allowed: bool
    choices: _ChoicesList | None
    db_column: str | None
    db_comment: str | None
    db_default: type[NOT_PROVIDED] | Expression
    column: str
    concrete: bool
    default: Any
    empty_values: Sequence[Any]
    creation_counter: int
    auto_creation_counter: int
    default_validators: Sequence[validators._ValidatorCallable]
    default_error_messages: ClassVar[_ErrorMessagesDict]
    hidden: bool
    system_check_removed_details: Any | None
    system_check_deprecated_details: Any | None
    non_db_attrs: tuple[str, ...]
    def __init__(
        self,
        verbose_name: _StrOrPromise | None = None,
        name: str | None = None,
        primary_key: bool = False,
        max_length: int | None = None,
        unique: bool = False,
        blank: bool = False,
        null: bool = False,
        db_index: bool = False,
        rel: ForeignObjectRel | None = None,
        default: Any = ...,
        editable: bool = True,
        serialize: bool = True,
        unique_for_date: str | None = None,
        unique_for_month: str | None = None,
        unique_for_year: str | None = None,
        choices: _ChoicesInput | None = None,
        help_text: _StrOrPromise = "",
        db_column: str | None = None,
        db_tablespace: str | None = None,
        auto_created: bool = False,
        validators: Iterable[validators._ValidatorCallable] = (),
        error_messages: _ErrorMessagesMapping | None = None,
        db_comment: str | None = None,
        db_default: type[NOT_PROVIDED] | Expression | _ST = ...,
    ) -> None: ...
    def __set__(self, instance: Any, value: _ST) -> None: ...
    # class access
    @overload
    def __get__(self, instance: None, owner: Any) -> _FieldDescriptor[Self]: ...
    # Model instance access
    @overload
    def __get__(self, instance: Model, owner: Any) -> _GT: ...
    # non-Model instances
    @overload
    def __get__(self, instance: Any, owner: Any) -> Self: ...
    def check(self, **kwargs: Any) -> list[CheckMessage]: ...
    def get_col(self, alias: str, output_field: Field | None = None) -> Col: ...
    @cached_property
    def cached_col(self) -> Col: ...
    def select_format(self, compiler: SQLCompiler, sql: str, params: _ParamsT) -> _AsSqlType: ...
    def deconstruct(self) -> tuple[str, str, Sequence[Any], dict[str, Any]]: ...
    def clone(self) -> Self: ...
    def __lt__(self, other: Any) -> bool: ...
    def __le__(self, other: Any) -> bool: ...
    def __gt__(self, other: Any) -> bool: ...
    def __ge__(self, other: Any) -> bool: ...
    def __deepcopy__(self, memodict: dict[int, Any]) -> Self: ...
    def __copy__(self) -> Self: ...
    def get_pk_value_on_save(self, instance: Model) -> Any: ...
    def to_python(self, value: Any) -> Any: ...
    @cached_property
    def error_messages(self) -> _ErrorMessagesDict: ...
    @cached_property
    def validators(self) -> list[validators._ValidatorCallable]: ...
    def run_validators(self, value: Any) -> None: ...
    def validate(self, value: Any, model_instance: Model | None) -> None: ...
    def clean(self, value: Any, model_instance: Model | None) -> Any: ...
    def db_type_parameters(self, connection: BaseDatabaseWrapper) -> DictWrapper: ...
    def db_check(self, connection: BaseDatabaseWrapper) -> str | None: ...
    def db_type(self, connection: BaseDatabaseWrapper) -> str | None: ...
    def rel_db_type(self, connection: BaseDatabaseWrapper) -> str | None: ...
    def cast_db_type(self, connection: BaseDatabaseWrapper) -> str | None: ...
    def db_parameters(self, connection: BaseDatabaseWrapper) -> dict[str, str | None]: ...
    def db_type_suffix(self, connection: BaseDatabaseWrapper) -> str | None: ...
    def get_db_converters(self, connection: BaseDatabaseWrapper) -> list[Callable[..., Any]]: ...
    @cached_property
    def unique(self) -> bool: ...
    @property
    def db_tablespace(self) -> str: ...
    @property
    def db_returning(self) -> bool: ...
    descriptor_class: type
    def set_attributes_from_name(self, name: str) -> None: ...
    def contribute_to_class(self, cls: type[Model], name: str, private_only: bool = False) -> None: ...
    def get_filter_kwargs_for_object(self, obj: Model) -> dict[str, Any]: ...
    def get_attname(self) -> str: ...
    def get_attname_column(self) -> tuple[str, str]: ...
    def get_internal_type(self) -> str: ...
    def pre_save(self, model_instance: Model, add: bool) -> Any: ...
    def get_prep_value(self, value: Any) -> Any: ...
    def get_db_prep_value(self, value: Any, connection: BaseDatabaseWrapper, prepared: bool = False) -> Any: ...
    def get_db_prep_save(self, value: Any, connection: BaseDatabaseWrapper) -> Any: ...
    def has_default(self) -> bool: ...
    def has_db_default(self) -> bool: ...
    def get_default(self) -> Any: ...
    def get_choices(
        self,
        include_blank: bool = True,
        blank_choice: _ChoicesList = ...,
        limit_choices_to: _LimitChoicesTo | None = None,
        ordering: Sequence[_OrderByFieldName] = (),
    ) -> BlankChoiceIterator | _ChoicesList: ...
    def value_to_string(self, obj: Model) -> str: ...
    @property
    def flatchoices(self) -> list[_Choice]: ...
    def save_form_data(self, instance: Model, data: Any) -> None: ...
    def formfield(
        self,
        form_class: type[forms.Field] | None = None,
        choices_form_class: type[forms.ChoiceField] | None = None,
        **kwargs: Any,
    ) -> forms.Field | None: ...
    def value_from_object(self, obj: Model) -> _GT: ...
    def slice_expression(self, expression: Expression, start: int, length: int | None) -> Func: ...

class IntegerField(Field[_ST, _GT]):
    _pyi_private_set_type: float | int | str | Combinable
    _pyi_private_get_type: int
    _pyi_lookup_exact_type: str | int
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class PositiveIntegerRelDbTypeMixin:
    def rel_db_type(self, connection: BaseDatabaseWrapper) -> str: ...

class SmallIntegerField(IntegerField[_ST, _GT]): ...

class BigIntegerField(IntegerField[_ST, _GT]):
    MAX_BIGINT: ClassVar[int]
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class PositiveIntegerField(PositiveIntegerRelDbTypeMixin, IntegerField[_ST, _GT]):
    integer_field_class: type[IntegerField]
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class PositiveSmallIntegerField(PositiveIntegerRelDbTypeMixin, SmallIntegerField[_ST, _GT]):
    integer_field_class: type[SmallIntegerField]
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class PositiveBigIntegerField(PositiveIntegerRelDbTypeMixin, BigIntegerField[_ST, _GT]):
    integer_field_class: type[BigIntegerField]
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class FloatField(Field[_ST, _GT]):
    _pyi_private_set_type: float | int | str | Combinable
    _pyi_private_get_type: float
    _pyi_lookup_exact_type: float
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class DecimalField(Field[_ST, _GT]):
    _pyi_private_set_type: str | float | decimal.Decimal | Combinable
    _pyi_private_get_type: decimal.Decimal
    _pyi_lookup_exact_type: str | int | decimal.Decimal
    # attributes
    max_digits: int
    decimal_places: int
    def __init__(
        self,
        verbose_name: _StrOrPromise | None = None,
        name: str | None = None,
        max_digits: int | None = None,
        decimal_places: int | None = None,
        *,
        primary_key: bool = ...,
        unique: bool = ...,
        blank: bool = ...,
        null: bool = ...,
        db_index: bool = ...,
        default: Any = ...,
        db_default: type[NOT_PROVIDED] | Expression | _ST = ...,
        editable: bool = ...,
        auto_created: bool = ...,
        serialize: bool = ...,
        choices: _ChoicesInput | None = ...,
        help_text: _StrOrPromise = ...,
        db_column: str | None = ...,
        db_comment: str | None = ...,
        db_tablespace: str | None = ...,
        validators: Iterable[validators._ValidatorCallable] = ...,
        error_messages: _ErrorMessagesMapping | None = ...,
    ) -> None: ...
    @cached_property
    def context(self) -> decimal.Context: ...
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class CharField(Field[_ST, _GT]):
    _pyi_private_set_type: str | int | Combinable
    _pyi_private_get_type: str
    # objects are converted to string before comparison
    _pyi_lookup_exact_type: Any
    def __init__(
        self,
        verbose_name: _StrOrPromise | None = ...,
        name: str | None = ...,
        primary_key: bool = ...,
        max_length: int | None = ...,
        unique: bool = ...,
        blank: bool = ...,
        null: bool = ...,
        db_index: bool = ...,
        default: Any = ...,
        db_default: type[NOT_PROVIDED] | Expression | _ST = ...,
        editable: bool = ...,
        auto_created: bool = ...,
        serialize: bool = ...,
        unique_for_date: str | None = ...,
        unique_for_month: str | None = ...,
        unique_for_year: str | None = ...,
        choices: _ChoicesInput | None = ...,
        help_text: _StrOrPromise = ...,
        db_column: str | None = ...,
        db_comment: str | None = ...,
        db_tablespace: str | None = ...,
        validators: Iterable[validators._ValidatorCallable] = ...,
        error_messages: _ErrorMessagesMapping | None = ...,
        *,
        db_collation: str | None = None,
    ) -> None: ...
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class CommaSeparatedIntegerField(CharField[_ST, _GT]): ...

class SlugField(CharField[_ST, _GT]):
    def __init__(
        self,
        verbose_name: _StrOrPromise | None = ...,
        name: str | None = ...,
        primary_key: bool = ...,
        unique: bool = ...,
        blank: bool = ...,
        null: bool = ...,
        default: Any = ...,
        db_default: type[NOT_PROVIDED] | Expression | _ST = ...,
        editable: bool = ...,
        auto_created: bool = ...,
        serialize: bool = ...,
        unique_for_date: str | None = ...,
        unique_for_month: str | None = ...,
        unique_for_year: str | None = ...,
        choices: _ChoicesInput | None = ...,
        help_text: _StrOrPromise = ...,
        db_column: str | None = ...,
        db_comment: str | None = ...,
        db_tablespace: str | None = ...,
        validators: Iterable[validators._ValidatorCallable] = ...,
        error_messages: _ErrorMessagesMapping | None = ...,
        *,
        max_length: int | None = 50,
        db_index: bool = True,
        allow_unicode: bool = False,
    ) -> None: ...
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class EmailField(CharField[_ST, _GT]):
    _pyi_private_set_type: str | Combinable
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class URLField(CharField[_ST, _GT]):
    def __init__(
        self,
        verbose_name: _StrOrPromise | None = None,
        name: str | None = None,
        *,
        primary_key: bool = ...,
        max_length: int | None = ...,
        unique: bool = ...,
        blank: bool = ...,
        null: bool = ...,
        db_index: bool = ...,
        rel: ForeignObjectRel | None = ...,
        default: Any = ...,
        db_default: type[NOT_PROVIDED] | Expression | _ST = ...,
        editable: bool = ...,
        serialize: bool = ...,
        unique_for_date: str | None = ...,
        unique_for_month: str | None = ...,
        unique_for_year: str | None = ...,
        choices: _ChoicesInput | None = ...,
        help_text: _StrOrPromise = ...,
        db_column: str | None = ...,
        db_comment: str | None = ...,
        db_tablespace: str | None = ...,
        auto_created: bool = ...,
        validators: Iterable[validators._ValidatorCallable] = ...,
        error_messages: _ErrorMessagesMapping | None = ...,
    ) -> None: ...
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class TextField(Field[_ST, _GT]):
    _pyi_private_set_type: str | Combinable
    _pyi_private_get_type: str
    # objects are converted to string before comparison
    _pyi_lookup_exact_type: Any
    def __init__(
        self,
        verbose_name: _StrOrPromise | None = ...,
        name: str | None = ...,
        primary_key: bool = ...,
        max_length: int | None = ...,
        unique: bool = ...,
        blank: bool = ...,
        null: bool = ...,
        db_index: bool = ...,
        default: Any = ...,
        db_default: type[NOT_PROVIDED] | Expression | _ST = ...,
        editable: bool = ...,
        auto_created: bool = ...,
        serialize: bool = ...,
        unique_for_date: str | None = ...,
        unique_for_month: str | None = ...,
        unique_for_year: str | None = ...,
        choices: _ChoicesInput | None = ...,
        help_text: _StrOrPromise = ...,
        db_column: str | None = ...,
        db_comment: str | None = ...,
        db_tablespace: str | None = ...,
        validators: Iterable[validators._ValidatorCallable] = ...,
        error_messages: _ErrorMessagesMapping | None = ...,
        *,
        db_collation: str | None = None,
    ) -> None: ...
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class BooleanField(Field[_ST, _GT]):
    _pyi_private_set_type: bool | Combinable
    _pyi_private_get_type: bool
    _pyi_lookup_exact_type: bool
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class NullBooleanField(BooleanField[_ST, _GT]):
    _pyi_private_set_type: bool | Combinable | None  # type: ignore[assignment]
    _pyi_private_get_type: bool | None  # type: ignore[assignment]
    _pyi_lookup_exact_type: bool | None  # type: ignore[assignment]

class IPAddressField(Field[_ST, _GT]):
    _pyi_private_set_type: str | Combinable
    _pyi_private_get_type: str

class GenericIPAddressField(Field[_ST, _GT]):
    _pyi_private_set_type: str | int | Callable[..., Any] | Combinable
    _pyi_private_get_type: str

    default_error_messages: ClassVar[_ErrorMessagesDict]
    unpack_ipv4: bool
    protocol: str
    def __init__(
        self,
        verbose_name: _StrOrPromise | None = None,
        name: Any | None = None,
        protocol: str = "both",
        unpack_ipv4: bool = False,
        primary_key: bool = ...,
        unique: bool = ...,
        blank: bool = ...,
        null: bool = ...,
        db_index: bool = ...,
        default: Any = ...,
        db_default: type[NOT_PROVIDED] | Expression | _ST = ...,
        editable: bool = ...,
        auto_created: bool = ...,
        serialize: bool = ...,
        choices: _ChoicesInput | None = ...,
        help_text: _StrOrPromise = ...,
        db_column: str | None = ...,
        db_comment: str | None = ...,
        db_tablespace: str | None = ...,
        validators: Iterable[validators._ValidatorCallable] = ...,
        error_messages: _ErrorMessagesMapping | None = ...,
    ) -> None: ...
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class DateTimeCheckMixin:
    def check(self, **kwargs: Any) -> list[CheckMessage]: ...

class DateField(DateTimeCheckMixin, Field[_ST, _GT]):
    _pyi_private_set_type: str | date | Combinable
    _pyi_private_get_type: date
    _pyi_lookup_exact_type: str | date
    auto_now: bool
    auto_now_add: bool
    def __init__(
        self,
        verbose_name: _StrOrPromise | None = None,
        name: str | None = None,
        auto_now: bool = False,
        auto_now_add: bool = False,
        *,
        primary_key: bool = ...,
        max_length: int | None = ...,
        unique: bool = ...,
        blank: bool = ...,
        null: bool = ...,
        db_index: bool = ...,
        default: Any = ...,
        db_default: type[NOT_PROVIDED] | Expression | _ST = ...,
        editable: bool = ...,
        auto_created: bool = ...,
        serialize: bool = ...,
        choices: _ChoicesInput | None = ...,
        help_text: _StrOrPromise = ...,
        db_column: str | None = ...,
        db_comment: str | None = ...,
        db_tablespace: str | None = ...,
        validators: Iterable[validators._ValidatorCallable] = ...,
        error_messages: _ErrorMessagesMapping | None = ...,
    ) -> None: ...
    @override
    def contribute_to_class(self, cls: type[Model], name: str, **kwargs: Any) -> None: ...  # type: ignore[override]
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class TimeField(DateTimeCheckMixin, Field[_ST, _GT]):
    _pyi_private_set_type: str | time | real_datetime | Combinable
    _pyi_private_get_type: time
    auto_now: bool
    auto_now_add: bool
    def __init__(
        self,
        verbose_name: _StrOrPromise | None = None,
        name: str | None = None,
        auto_now: bool = False,
        auto_now_add: bool = False,
        *,
        primary_key: bool = ...,
        unique: bool = ...,
        blank: bool = ...,
        null: bool = ...,
        db_index: bool = ...,
        default: Any = ...,
        db_default: type[NOT_PROVIDED] | Expression | _ST = ...,
        editable: bool = ...,
        auto_created: bool = ...,
        serialize: bool = ...,
        choices: _ChoicesInput | None = ...,
        help_text: _StrOrPromise = ...,
        db_column: str | None = ...,
        db_comment: str | None = ...,
        db_tablespace: str | None = ...,
        validators: Iterable[validators._ValidatorCallable] = ...,
        error_messages: _ErrorMessagesMapping | None = ...,
    ) -> None: ...
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class DateTimeField(DateField[_ST, _GT]):
    _pyi_private_set_type: str | real_datetime | date | Combinable
    _pyi_private_get_type: real_datetime
    _pyi_lookup_exact_type: str | real_datetime
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class UUIDField(Field[_ST, _GT]):
    _pyi_private_set_type: str | uuid.UUID
    _pyi_private_get_type: uuid.UUID
    _pyi_lookup_exact_type: uuid.UUID | str
    def __init__(
        self,
        verbose_name: _StrOrPromise | None = None,
        *,
        name: str | None = ...,
        primary_key: bool = ...,
        max_length: int | None = ...,
        unique: bool = ...,
        blank: bool = ...,
        null: bool = ...,
        db_index: bool = ...,
        rel: ForeignObjectRel | None = ...,
        default: Any = ...,
        db_default: type[NOT_PROVIDED] | Expression | _ST = ...,
        editable: bool = ...,
        serialize: bool = ...,
        unique_for_date: str | None = ...,
        unique_for_month: str | None = ...,
        unique_for_year: str | None = ...,
        choices: _ChoicesInput | None = ...,
        help_text: _StrOrPromise = ...,
        db_column: str | None = ...,
        db_comment: str | None = ...,
        db_tablespace: str | None = ...,
        auto_created: bool = ...,
        validators: Iterable[validators._ValidatorCallable] = ...,
        error_messages: _ErrorMessagesMapping | None = ...,
    ) -> None: ...
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class FilePathField(Field[_ST, _GT]):
    path: Any
    match: str | None
    recursive: bool
    allow_files: bool
    allow_folders: bool
    def __init__(
        self,
        verbose_name: _StrOrPromise | None = None,
        name: str | None = None,
        path: str | Callable[..., str] = "",
        match: str | None = None,
        recursive: bool = False,
        allow_files: bool = True,
        allow_folders: bool = False,
        *,
        primary_key: bool = ...,
        max_length: int = ...,
        unique: bool = ...,
        blank: bool = ...,
        null: bool = ...,
        db_index: bool = ...,
        default: Any = ...,
        db_default: type[NOT_PROVIDED] | Expression | _ST = ...,
        editable: bool = ...,
        auto_created: bool = ...,
        serialize: bool = ...,
        choices: _ChoicesInput | None = ...,
        help_text: _StrOrPromise = ...,
        db_column: str | None = ...,
        db_comment: str | None = ...,
        db_tablespace: str | None = ...,
        validators: Iterable[validators._ValidatorCallable] = ...,
        error_messages: _ErrorMessagesMapping | None = ...,
    ) -> None: ...
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class BinaryField(Field[_ST, _GT]):
    _pyi_private_get_type: bytes | memoryview
    def get_placeholder(self, value: Any, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> str: ...

class DurationField(Field[_ST, _GT]):
    _pyi_private_get_type: timedelta
    @override
    def formfield(self, **kwargs: Any) -> forms.Field | None: ...  # type: ignore[override]

class AutoFieldMixin:
    db_returning: bool
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def check(self, **kwargs: Any) -> list[CheckMessage]: ...
    def deconstruct(self) -> tuple[str, str, Sequence[Any], dict[str, Any]]: ...
    def validate(self, value: Any, model_instance: Model | None) -> None: ...
    def get_db_prep_value(self, value: Any, connection: BaseDatabaseWrapper, prepared: bool = False) -> Any: ...
    def contribute_to_class(self, cls: type[Model], name: str, **kwargs: Any) -> None: ...
    def formfield(self, **kwargs: Any) -> None: ...

class AutoFieldMeta(type): ...

class AutoField(AutoFieldMixin, IntegerField[_ST, _GT], metaclass=AutoFieldMeta):  # type: ignore[misc]
    _pyi_private_set_type: Combinable | int | str
    _pyi_private_get_type: int
    _pyi_lookup_exact_type: str | int

class BigAutoField(AutoFieldMixin, BigIntegerField[_ST, _GT]): ...  # type: ignore[misc]
class SmallAutoField(AutoFieldMixin, SmallIntegerField[_ST, _GT]): ...  # type: ignore[misc]

__all__ = [
    "BLANK_CHOICE_DASH",
    "NOT_PROVIDED",
    "AutoField",
    "BigAutoField",
    "BigIntegerField",
    "BinaryField",
    "BooleanField",
    "CharField",
    "CommaSeparatedIntegerField",
    "DateField",
    "DateTimeField",
    "DecimalField",
    "DurationField",
    "EmailField",
    "Empty",
    "Field",
    "FilePathField",
    "FloatField",
    "GenericIPAddressField",
    "IPAddressField",
    "IntegerField",
    "NullBooleanField",
    "PositiveBigIntegerField",
    "PositiveIntegerField",
    "PositiveSmallIntegerField",
    "SlugField",
    "SmallAutoField",
    "SmallIntegerField",
    "TextField",
    "TimeField",
    "URLField",
    "UUIDField",
]



from collections.abc import Iterable
from typing import Any

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models.lookups import Exact, GreaterThan, GreaterThanOrEqual, In, IsNull, LessThan, LessThanOrEqual
from django.db.models.sql.compiler import SQLCompiler, _AsSqlType
from typing_extensions import override

def get_normalized_value(value: Any, lhs: Any) -> tuple[Any, ...]: ...

class RelatedIn(In):
    @override
    def get_prep_lookup(self) -> Iterable[Any]: ...

class RelatedLookupMixin:
    rhs: Any
    def get_prep_lookup(self) -> Any: ...
    def as_sql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...

class RelatedExact(RelatedLookupMixin, Exact): ...
class RelatedLessThan(RelatedLookupMixin, LessThan): ...
class RelatedGreaterThan(RelatedLookupMixin, GreaterThan): ...
class RelatedGreaterThanOrEqual(RelatedLookupMixin, GreaterThanOrEqual): ...
class RelatedLessThanOrEqual(RelatedLookupMixin, LessThanOrEqual): ...
class RelatedIsNull(RelatedLookupMixin, IsNull): ...



from pathlib import Path
from typing import Any

def get_template_directories() -> set[Path]: ...
def reset_loaders() -> None: ...
def watch_for_template_changes(sender: object, **kwargs: Any) -> None: ...
def template_changed(sender: object | None, file_path: Path, **kwargs: Any) -> bool | None: ...



from django.db.models.base import Model
from mypy.errorcodes import CALL_ARG
from mypy.plugin import FunctionContext, MethodContext
from mypy.types import Instance, get_proper_type
from mypy.types import Type as MypyType

from mypy_django_plugin.django.context import DjangoContext
from mypy_django_plugin.lib import helpers


def get_actual_types(ctx: MethodContext | FunctionContext, expected_keys: list[str]) -> list[tuple[str, MypyType]]:
    actual_types = []
    # positionals
    for pos, (actual_name, actual_type) in enumerate(zip(ctx.arg_names[0], ctx.arg_types[0], strict=False)):
        if actual_name is None:
            if ctx.callee_arg_names[0] == "kwargs" or pos >= len(expected_keys):
                # unpacked dict as kwargs is not supported
                continue
            actual_name = expected_keys[pos]
        actual_types.append((actual_name, actual_type))
    # kwargs
    if len(ctx.callee_arg_names) > 1:
        for actual_name, actual_type in zip(ctx.arg_names[1], ctx.arg_types[1], strict=False):
            if actual_name is None:
                # unpacked dict as kwargs is not supported
                continue
            actual_types.append((actual_name, actual_type))
    return actual_types


def typecheck_model_method(
    ctx: FunctionContext | MethodContext,
    django_context: DjangoContext,
    model_cls: type[Model],
    method: str,
) -> None:
    """Type-checks positional and keyword arguments for Model methods like __init__(), create(), and acreate()."""
    typechecker_api = helpers.get_typechecker_api(ctx)
    expected_types = django_context.get_expected_types(typechecker_api, model_cls, method=method)
    expected_keys = [key for key in expected_types.keys() if key != "pk"]

    min_arg_count = helpers.get_min_argument_count(ctx)

    for actual_name, actual_type in get_actual_types(ctx, expected_keys):
        if actual_name not in expected_types:
            ctx.api.fail(f'Unexpected attribute "{actual_name}" for model "{model_cls.__name__}"', ctx.context)
            min_arg_count -= 1  # To avoid double error (Unexpected attribute + too many arguments)
            continue
        helpers.check_types_compatible(
            ctx,
            expected_type=expected_types[actual_name],
            actual_type=actual_type,
            error_message=f'Incompatible type for "{actual_name}" of "{model_cls.__name__}"',
        )

    if min_arg_count > len(expected_keys):
        ctx.api.fail(f'Too many arguments for "{model_cls.__name__}"', ctx.context, code=CALL_ARG)


def typecheck_model_init(ctx: FunctionContext, django_context: DjangoContext) -> MypyType:
    default_return_type = get_proper_type(ctx.default_return_type)
    if (
        isinstance(default_return_type, Instance)
        and (model_cls := django_context.get_model_class_by_fullname(default_return_type.type.fullname)) is not None
    ):
        typecheck_model_method(ctx, django_context, model_cls, "__init__")

    return ctx.default_return_type


def typecheck_model_create(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    default_return_type = get_proper_type(ctx.default_return_type)
    if (
        isinstance(default_return_type, Instance)
        and (model_cls := django_context.get_model_class_by_fullname(default_return_type.type.fullname)) is not None
    ):
        typecheck_model_method(ctx, django_context, model_cls, "create")

    return ctx.default_return_type


def typecheck_model_acreate(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    default_return_type = get_proper_type(ctx.default_return_type)
    if (
        isinstance(default_return_type, Instance)
        # default_return_type at this point should be of type Coroutine[Any, Any, <Model>]
        and isinstance((model := get_proper_type(default_return_type.args[-1])), Instance)
        and (model_cls := django_context.get_model_class_by_fullname(model.type.fullname)) is not None
    ):
        typecheck_model_method(ctx, django_context, model_cls, "acreate")

    return ctx.default_return_type



from typing import Any

from django.core.checks import CheckMessage
from django.core.exceptions import ValidationError

def prefix_validation_error(error: Any, prefix: Any, code: Any, params: Any) -> ValidationError: ...

class CheckPostgresInstalledMixin:
    def _check_postgres_installed(self, *args: Any) -> list[CheckMessage]: ...
    def check(self, *args: Any, **kwargs: Any) -> list[CheckMessage]: ...



from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.backends.sqlite3.base import DatabaseWrapper
from django.db.models.base import Model
from typing_extensions import override

class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    connection: DatabaseWrapper
    sql_create_fk: None  # type: ignore[assignment]
    sql_alter_table_comment: None  # type: ignore[assignment]
    sql_alter_column_comment: None  # type: ignore[assignment]
    @override
    def delete_model(self, model: type[Model], handle_autom2m: bool = True) -> None: ...



from collections.abc import Callable, Collection, Mapping
from typing import Any, TypeVar

_F = TypeVar("_F", bound=Callable[..., Any])

def csp_override(config: Mapping[str, Collection[str] | str]) -> Callable[[_F], _F]: ...
def csp_report_only_override(config: Mapping[str, Collection[str] | str]) -> Callable[[_F], _F]: ...



from collections.abc import Iterable
from typing import Any, ClassVar, Literal, overload
from uuid import UUID

from django.db import models
from django.db.models.base import Model

ADDITION: int
CHANGE: int
DELETION: int
ACTION_FLAG_CHOICES: Any

class LogEntryManager(models.Manager[LogEntry]):
    @overload
    def log_actions(
        self,
        user_id: int | str | UUID,
        queryset: Iterable[Model],
        action_flag: int,
        change_message: str | list[Any] = "",
        *,
        single_object: Literal[True],
    ) -> LogEntry: ...
    @overload
    def log_actions(
        self,
        user_id: int | str | UUID,
        queryset: Iterable[Model],
        action_flag: int,
        change_message: str | list[Any] = "",
        *,
        single_object: Literal[False] = ...,
    ) -> list[LogEntry]: ...

class LogEntry(models.Model):
    action_time: models.DateTimeField
    user: models.ForeignKey
    content_type: models.ForeignKey
    object_id: models.TextField
    object_repr: models.CharField
    action_flag: models.PositiveSmallIntegerField
    change_message: models.TextField
    objects: ClassVar[LogEntryManager]
    def is_addition(self) -> bool: ...
    def is_change(self) -> bool: ...
    def is_deletion(self) -> bool: ...
    def get_change_message(self) -> str: ...
    def get_edited_object(self) -> Model: ...
    def get_admin_url(self) -> str | None: ...



from typing import Any, Generic, TypeVar

from django.contrib.sessions.backends.base import SessionBase
from django.contrib.sessions.base_session import AbstractBaseSession
from django.utils.functional import cached_property

_ST = TypeVar("_ST", bound=AbstractBaseSession, default=AbstractBaseSession)

class SessionStore(SessionBase, Generic[_ST]):
    def __init__(self, session_key: str | None = None) -> None: ...
    @classmethod
    def get_model_class(cls) -> type[_ST]: ...
    @cached_property
    def model(self) -> type[_ST]: ...
    def create_model_instance(self, data: dict[str, Any]) -> _ST: ...
    async def acreate_model_instance(self, data: dict[str, Any]) -> _ST: ...



from logging import Logger
from typing import Any

from django.tasks.base import Task, TaskResult
from typing_extensions import override

from .base import BaseTaskBackend

logger: Logger

class ImmediateBackend(BaseTaskBackend):
    worker_id: str
    def __init__(self, alias: str, params: dict[str, Any]) -> None: ...
    @override
    def enqueue(self, task: Task, args: list[Any], kwargs: dict[str, Any]) -> TaskResult: ...



from typing import TypeAlias

PYPY: bool

PY38: bool
PY39: bool
PY310: bool
PY311: bool
PY312: bool
PY313: bool
PY314: bool

_VT: TypeAlias = tuple[int, int, int, str, int]

def get_version(version: _VT | None = None) -> str: ...
def get_main_version(version: _VT | None = None) -> str: ...
def get_complete_version(version: _VT | None = None) -> _VT: ...
def get_docs_version(version: _VT | None = None) -> str: ...
def get_git_changeset() -> str | None: ...
def get_version_tuple(version: str) -> tuple[int, int, int]: ...



from collections.abc import Callable
from hmac import HMAC

RANDOM_STRING_CHARS: str

class InvalidAlgorithm(ValueError): ...

def salted_hmac(
    key_salt: bytes | str, value: bytes | str, secret: bytes | str | None = None, *, algorithm: str = "sha1"
) -> HMAC: ...
def get_random_string(length: int, allowed_chars: str = ...) -> str: ...
def constant_time_compare(val1: bytes | str, val2: bytes | str) -> bool: ...
def pbkdf2(
    password: bytes | str,
    salt: bytes | str,
    iterations: int,
    dklen: int = 0,
    digest: Callable | None = None,
) -> bytes: ...



from typing import Any

from django.core.checks.messages import CheckMessage
from django.db.models.base import Model
from django.utils.functional import cached_property

NOT_PROVIDED: Any

class FieldCacheMixin:
    @cached_property
    def cache_name(self) -> str: ...
    def get_cached_value(self, instance: Model, default: Any = ...) -> Model | None: ...
    def is_cached(self, instance: Model) -> bool: ...
    def set_cached_value(self, instance: Model, value: Model | None) -> None: ...
    def delete_cached_value(self, instance: Model) -> None: ...

class CheckFieldDefaultMixin:
    def check(self, **kwargs: Any) -> list[CheckMessage]: ...



from _collections_abc import dict_items, dict_keys, dict_values
from collections.abc import Iterable
from datetime import datetime, timedelta
from typing import Any, overload

from _typeshed import SupportsKeysAndGetItem
from typing_extensions import override

VALID_KEY_CHARS: Any

class CreateError(Exception): ...
class UpdateError(Exception): ...

class SessionBase(dict[str, Any]):
    TEST_COOKIE_NAME: str
    TEST_COOKIE_VALUE: str
    accessed: bool
    modified: bool
    serializer: Any
    def __init__(self, session_key: str | None = None) -> None: ...
    async def aset(self, key: str, value: Any) -> None: ...
    @property
    def key_salt(self) -> str: ...
    @overload  # type: ignore[override]
    def get(self, key: str) -> Any | None: ...
    @overload
    @override
    def get(self, key: str, default: Any) -> Any: ...
    @overload
    async def aget(self, key: str) -> Any | None: ...
    @overload
    async def aget(self, key: str, default: Any) -> Any: ...
    @overload
    @override
    def pop(self, key: str) -> Any: ...
    @overload
    @override
    def pop(self, key: str, default: Any) -> Any: ...
    @overload
    async def apop(self, key: str) -> Any: ...
    @overload
    async def apop(self, key: str, default: Any) -> Any: ...
    @override
    def setdefault(self, key: str, value: Any) -> Any: ...  # type: ignore[override]
    async def asetdefault(self, key: str, value: Any) -> Any: ...
    def set_test_cookie(self) -> None: ...
    async def aset_test_cookie(self) -> None: ...
    def test_cookie_worked(self) -> bool: ...
    async def atest_cookie_worked(self) -> bool: ...
    def delete_test_cookie(self) -> None: ...
    async def adelete_test_cookie(self) -> None: ...
    def encode(self, session_dict: dict[str, Any]) -> str: ...
    def decode(self, session_data: bytes | str) -> dict[str, Any]: ...
    @overload  # type: ignore[override]
    def update(self, dict_: SupportsKeysAndGetItem[str, Any]) -> None: ...
    @overload
    @override
    def update(self, dict_: Iterable[tuple[str, Any]]) -> None: ...
    @overload
    async def aupdate(self, dict_: SupportsKeysAndGetItem[str, Any]) -> None: ...
    @overload
    async def aupdate(self, dict_: Iterable[tuple[str, Any]]) -> None: ...
    def has_key(self, key: Any) -> bool: ...
    async def ahas_key(self, key: Any) -> bool: ...
    async def akeys(self) -> dict_keys[str, Any]: ...
    async def avalues(self) -> dict_values[str, Any]: ...
    async def aitems(self) -> dict_items[str, Any]: ...
    @override
    def clear(self) -> None: ...
    def is_empty(self) -> bool: ...
    def _get_session_key(self) -> str | None: ...
    def _set_session_key(self, value: str | None) -> None: ...
    @property
    def session_key(self) -> str | None: ...
    @property
    def _session_key(self) -> str | None: ...
    @_session_key.setter
    def _session_key(self, value: str | None) -> None: ...
    def get_session_cookie_age(self) -> int: ...
    def get_expiry_age(self, **kwargs: Any) -> int: ...
    async def aget_expiry_age(self, **kwargs: Any) -> int: ...
    def get_expiry_date(self, **kwargs: Any) -> datetime: ...
    async def aget_expiry_date(self, **kwargs: Any) -> datetime: ...
    def set_expiry(self, value: datetime | timedelta | int | None) -> None: ...
    async def aset_expiry(self, value: datetime | timedelta | int | None) -> None: ...
    def get_expire_at_browser_close(self) -> bool: ...
    async def aget_expire_at_browser_close(self) -> bool: ...
    def flush(self) -> None: ...
    async def aflush(self) -> None: ...
    def cycle_key(self) -> None: ...
    async def acycle_key(self) -> None: ...
    def exists(self, session_key: str) -> bool: ...
    async def aexists(self, session_key: str) -> bool: ...
    def create(self) -> None: ...
    async def acreate(self) -> None: ...
    def save(self, must_create: bool = False) -> None: ...
    async def asave(self, must_create: bool = False) -> None: ...
    def delete(self, session_key: str | None = None) -> None: ...
    async def adelete(self, session_key: str | None = None) -> None: ...
    def load(self) -> dict[str, Any]: ...
    async def aload(self) -> dict[str, Any]: ...
    @classmethod
    def clear_expired(cls) -> None: ...
    @classmethod
    async def aclear_expired(cls) -> None: ...



from functools import cached_property
from typing import Any

from django.db.backends.oracle.introspection import DatabaseIntrospection
from typing_extensions import override

class OracleIntrospection(DatabaseIntrospection):
    @cached_property
    @override
    def data_types_reverse(self) -> dict[int, str]: ...  # type: ignore[override]
    def get_geometry_type(self, table_name: Any, description: Any) -> Any: ...



from typing import NamedTuple

from django.db.backends.base.introspection import BaseDatabaseIntrospection
from django.db.backends.postgresql.base import DatabaseWrapper

class FieldInfo(NamedTuple):
    name: str
    type_code: int
    display_size: int | None
    internal_size: int | None
    precision: int | None
    scale: int | None
    null_ok: bool
    default: str | None
    collation: str | None
    is_autofield: bool
    comment: str | None

class TableInfo(NamedTuple):
    name: str
    type: str
    comment: str | None

class DatabaseIntrospection(BaseDatabaseIntrospection):
    connection: DatabaseWrapper
    data_types_reverse: dict[int, str]
    index_default_access_method: str
    ignored_tables: list[str]



from collections.abc import Callable
from typing import Any

from django.http.response import HttpResponseBase
from django.urls.resolvers import URLPattern

def static(prefix: str, view: Callable[..., HttpResponseBase] = ..., **kwargs: Any) -> list[URLPattern]: ...



from types import ModuleType
from typing import Any

from django.db.backends.base.operations import BaseDatabaseOperations
from django.db.backends.postgresql.base import DatabaseWrapper

def get_json_dumps(encoder: type[Any] | None) -> Any: ...

class DatabaseOperations(BaseDatabaseOperations):
    connection: DatabaseWrapper
    explain_options: frozenset[str]
    integerfield_type_map: dict[str, type[Any]]
    numeric: ModuleType

    def compose_sql(self, sql: Any, params: Any) -> Any: ...



import sys
from collections.abc import Collection, Mapping

from django.utils.functional import SimpleLazyObject
from typing_extensions import override

if sys.version_info >= (3, 11):
    from enum import StrEnum as _StrEnum
else:
    from enum import Enum

    class _ReprEnum(Enum): ...  # type: ignore[misc]
    class _StrEnum(str, _ReprEnum): ...  # type: ignore[misc]

class CSP(_StrEnum):
    HEADER_ENFORCE = "Content-Security-Policy"
    HEADER_REPORT_ONLY = "Content-Security-Policy-Report-Only"

    NONE = "'none'"
    REPORT_SAMPLE = "'report-sample'"
    SELF = "'self'"
    STRICT_DYNAMIC = "'strict-dynamic'"
    UNSAFE_EVAL = "'unsafe-eval'"
    UNSAFE_HASHES = "'unsafe-hashes'"
    UNSAFE_INLINE = "'unsafe-inline'"
    WASM_UNSAFE_EVAL = "'wasm-unsafe-eval'"

    NONCE = "<CSP_NONCE_SENTINEL>"

class LazyNonce(SimpleLazyObject):
    def __init__(self) -> None: ...
    @override
    def __bool__(self) -> bool: ...

def generate_nonce() -> str: ...
def build_policy(config: Mapping[str, Collection[str] | str], nonce: SimpleLazyObject | str | None = None) -> str: ...



from typing import Any

from django.core.checks.messages import CheckMessage
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models.fields import Field

class BaseDatabaseValidation:
    connection: BaseDatabaseWrapper
    def __init__(self, connection: BaseDatabaseWrapper) -> None: ...
    def __del__(self) -> None: ...
    def check(self, **kwargs: Any) -> list[CheckMessage]: ...
    def check_field(self, field: Field, **kwargs: Any) -> list[CheckMessage]: ...



from typing import Any

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models.sql.compiler import SQLCompiler, _AsSqlType

class FixDecimalInputMixin:
    def as_postgresql(
        self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any
    ) -> _AsSqlType: ...

class FixDurationInputMixin:
    def as_mysql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any) -> _AsSqlType: ...
    def as_oracle(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, **extra_context: Any) -> _AsSqlType: ...

class NumericOutputFieldMixin: ...



from django.http import HttpRequest, HttpResponse
from django.utils.csp import CSP as CSP
from django.utils.csp import LazyNonce
from django.utils.deprecation import MiddlewareMixin

def get_nonce(request: HttpRequest) -> LazyNonce | None: ...

class ContentSecurityPolicyMiddleware(MiddlewareMixin):
    def process_request(self, request: HttpRequest) -> None: ...
    def process_response(self, request: HttpRequest, response: HttpResponse) -> HttpResponse: ...



from typing import Any, TypeAlias

_JsonPrimitive: TypeAlias = str | int | float | bool | None
_JsonValue: TypeAlias = _JsonPrimitive | list[Any] | dict[str, Any]

def normalize_json(obj: Any) -> _JsonValue: ...



from django.core.checks.messages import CheckMessage
from django.db.backends.base.validation import BaseDatabaseValidation
from django.db.backends.oracle.base import DatabaseWrapper
from django.db.models.fields import Field

class DatabaseValidation(BaseDatabaseValidation):
    connection: DatabaseWrapper
    def check_field_type(self, field: Field, field_type: str) -> list[CheckMessage]: ...



from ipaddress import IPv4Address, IPv6Address
from typing import Final, TypeAlias, TypedDict

from _typeshed import StrPath
from django.contrib.gis.geos import Point
from django.utils.functional import cached_property

HAS_GEOIP2: bool
SUPPORTED_DATABASE_TYPES: set[str]

_CoordType: TypeAlias = tuple[float, float] | tuple[None, None]
_QueryType: TypeAlias = str | IPv4Address | IPv6Address

class _CityResponse(TypedDict):
    accuracy_radius: int | None
    city: str
    continent_code: str
    continent_name: str
    country_code: str
    country_name: str
    is_in_european_union: bool
    latitude: float
    longitude: float
    metro_code: int | None
    postal_code: str | None
    region_code: str | None
    region_name: str | None
    time_zone: str | None
    # Kept for backward compatibility.
    dma_code: int | None
    region: str | None

class _CountryResponse(TypedDict):
    continent_code: str
    continent_name: str
    country_code: str
    country_name: str
    is_in_european_union: bool

class GeoIP2Exception(Exception): ...

class GeoIP2:
    MODE_AUTO: Final = 0
    MODE_MMAP_EXT: Final = 1
    MODE_MMAP: Final = 2
    MODE_FILE: Final = 4
    MODE_MEMORY: Final = 8
    cache_options: Final[frozenset[int]]

    def __init__(
        self,
        path: StrPath | None = None,
        cache: int = 0,
        country: str | None = None,
        city: str | None = None,
    ) -> None: ...
    def __del__(self) -> None: ...
    @cached_property
    def is_city(self) -> bool: ...
    @cached_property
    def is_country(self) -> bool: ...
    def city(self, query: _QueryType) -> _CityResponse: ...
    def country_code(self, query: _QueryType) -> str: ...
    def country_name(self, query: _QueryType) -> str: ...
    def country(self, query: _QueryType) -> _CountryResponse: ...
    def lon_lat(self, query: _QueryType) -> _CoordType: ...
    def lat_lon(self, query: _QueryType) -> _CoordType: ...
    def geos(self, query: _QueryType) -> Point: ...

__all__ = ["HAS_GEOIP2"]



from collections.abc import Sequence
from typing import TYPE_CHECKING, Literal

from django.core.exceptions import FieldDoesNotExist, FieldError
from django.db.models.base import Model
from django.db.models.constants import LOOKUP_SEP
from django.db.models.fields.related import RelatedField
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.db.models.fields.reverse_related import ForeignObjectRel
from django.db.models.lookups import Transform
from django.db.models.sql.query import Query
from mypy.checker import TypeChecker
from mypy.errorcodes import NO_REDEF
from mypy.nodes import ARG_NAMED, ARG_NAMED_OPT, ARG_STAR, CallExpr, Expression, ListExpr, SetExpr, TupleExpr
from mypy.plugin import FunctionContext, MethodContext
from mypy.types import AnyType, Instance, LiteralType, ProperType, TupleType, TypedDictType, TypeOfAny, get_proper_type
from mypy.types import Type as MypyType

from mypy_django_plugin.django.context import DjangoContext, LookupsAreUnsupported
from mypy_django_plugin.lib import fullnames, helpers
from mypy_django_plugin.lib.helpers import DjangoModel
from mypy_django_plugin.transformers.models import get_annotated_type

if TYPE_CHECKING:
    from django.db.models.options import _AnyField


def determine_proper_manager_type(ctx: FunctionContext) -> MypyType:
    default_return_type = get_proper_type(ctx.default_return_type)
    assert isinstance(default_return_type, Instance)

    outer_model_info = helpers.get_typechecker_api(ctx).scope.active_class()
    if (
        outer_model_info is None
        or not helpers.is_model_type(outer_model_info)
        or outer_model_info.self_type is None
        or not default_return_type.type.is_generic()
    ):
        return default_return_type

    return default_return_type.copy_modified(args=[outer_model_info.self_type])


def get_field_type_from_lookup(
    ctx: MethodContext,
    django_context: DjangoContext,
    model_cls: type[Model],
    *,
    method: str,
    lookup: str,
    silent_on_error: bool = False,
) -> MypyType | None:
    try:
        lookup_field, model_cls = django_context.resolve_lookup_into_field(model_cls, lookup)
    except FieldError as exc:
        if not silent_on_error:
            ctx.api.fail(exc.args[0], ctx.context)
        return None
    except LookupsAreUnsupported:
        return AnyType(TypeOfAny.explicit)

    if lookup_field is None:
        return AnyType(TypeOfAny.implementation_artifact)
    if (isinstance(lookup_field, RelatedField) and lookup_field.column == lookup) or isinstance(
        lookup_field, ForeignObjectRel
    ):
        model_cls = django_context.get_field_related_model_cls(lookup_field)
        lookup_field = django_context.get_primary_key_field(model_cls)

    api = helpers.get_typechecker_api(ctx)
    model_info = helpers.lookup_class_typeinfo(api, model_cls)
    return django_context.get_field_get_type(api, model_info, lookup_field, method=method)


def get_values_list_row_type(
    ctx: MethodContext,
    django_context: DjangoContext,
    model_cls: type[Model],
    *,
    is_annotated: bool,
    flat: bool,
    named: bool,
) -> MypyType:
    field_lookups = resolve_field_lookups(ctx.args[0], django_context)
    if field_lookups is None:
        return AnyType(TypeOfAny.from_error)

    typechecker_api = helpers.get_typechecker_api(ctx)
    model_info = helpers.lookup_class_typeinfo(typechecker_api, model_cls)
    if len(field_lookups) == 0:
        if flat:
            primary_key_field = django_context.get_primary_key_field(model_cls)
            lookup_type = get_field_type_from_lookup(
                ctx, django_context, model_cls, lookup=primary_key_field.attname, method="values_list"
            )
            assert lookup_type is not None
            return lookup_type
        if named:
            column_types: dict[str, MypyType] = {}
            for field in django_context.get_model_fields(model_cls):
                column_type = django_context.get_field_get_type(
                    typechecker_api, model_info, field, method="values_list"
                )
                column_types[field.attname] = column_type
            if is_annotated:
                # Return a NamedTuple with a fallback so that it's possible to access any field
                return helpers.make_oneoff_named_tuple(
                    typechecker_api,
                    "Row",
                    column_types,
                    extra_bases=[typechecker_api.named_generic_type(fullnames.ANY_ATTR_ALLOWED_CLASS_FULLNAME, [])],
                )
            return helpers.make_oneoff_named_tuple(typechecker_api, "Row", column_types)
        # flat=False, named=False, all fields
        if is_annotated:
            return typechecker_api.named_generic_type("builtins.tuple", [AnyType(TypeOfAny.special_form)])
        field_lookups = []
        for field in django_context.get_model_fields(model_cls):
            field_lookups.append(field.attname)

    if len(field_lookups) > 1 and flat:
        typechecker_api.fail("'flat' is not valid when 'values_list' is called with more than one field", ctx.context)
        return AnyType(TypeOfAny.from_error)

    column_types = {}
    for field_lookup in field_lookups:
        lookup_field_type = get_field_type_from_lookup(
            ctx, django_context, model_cls, lookup=field_lookup, method="values_list", silent_on_error=is_annotated
        )
        if lookup_field_type is None:
            if is_annotated:
                lookup_field_type = AnyType(TypeOfAny.from_omitted_generics)
            else:
                return AnyType(TypeOfAny.from_error)
        column_types[field_lookup] = lookup_field_type

    if flat:
        assert len(column_types) == 1
        row_type = next(iter(column_types.values()))
    elif named:
        row_type = helpers.make_oneoff_named_tuple(typechecker_api, "Row", column_types)
    else:
        # Since there may have been repeated field lookups, we cannot just use column_types.values here.
        # This is not the case in named above, because Django will error if duplicate fields are requested.
        resolved_column_types = [column_types[field_lookup] for field_lookup in field_lookups]
        row_type = helpers.make_tuple(typechecker_api, resolved_column_types)

    return row_type


def extract_proper_type_queryset_values_list(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    django_model = helpers.get_model_info_from_qs_ctx(ctx, django_context)
    if django_model is None:
        return ctx.default_return_type

    default_return_type = get_proper_type(ctx.default_return_type)
    if not isinstance(default_return_type, Instance):
        return ctx.default_return_type

    flat = helpers.get_bool_call_argument_by_name(ctx, "flat", default=False)
    named = helpers.get_bool_call_argument_by_name(ctx, "named", default=False)

    if flat and named:
        ctx.api.fail("'flat' and 'named' can't be used together", ctx.context)
        return default_return_type.copy_modified(args=[django_model.typ, AnyType(TypeOfAny.from_error)])

    row_type = get_values_list_row_type(
        ctx, django_context, django_model.cls, is_annotated=django_model.is_annotated, flat=flat, named=named
    )
    ret = default_return_type.copy_modified(args=[django_model.typ, row_type])
    if not named and (field_lookups := resolve_field_lookups(ctx.args[0], django_context)):
        # For non-named values_list, the row type does not encode column names.
        # Attach selected field names to the returned QuerySet instance so that
        # subsequent annotate() can make an informed decision about name conflicts.
        ret.extra_attrs = helpers.merge_extra_attrs(ret.extra_attrs, new_immutable=set(field_lookups))
    return ret


def gather_kwargs(ctx: MethodContext) -> dict[str, MypyType] | None:
    num_args = len(ctx.arg_kinds)
    kwargs = {}
    named = (ARG_NAMED, ARG_NAMED_OPT)
    for i in range(num_args):
        if not ctx.arg_kinds[i]:
            continue
        if any(kind not in named for kind in ctx.arg_kinds[i]):
            # Only named arguments supported
            continue
        for j in range(len(ctx.arg_names[i])):
            name = ctx.arg_names[i][j]
            assert name is not None
            kwargs[name] = ctx.arg_types[i][j]
    return kwargs


def gather_expression_types(ctx: MethodContext) -> dict[str, MypyType]:
    kwargs = gather_kwargs(ctx)
    if not kwargs:
        return {}

    # For now, we don't try to resolve the output_field of the field would be, but use Any.
    # NOTE: It's important that we use 'special_form' for 'Any' as otherwise we can
    # get stuck with mypy interpreting an overload ambiguity towards the
    # overloaded 'Field.__get__' method when its 'model' argument gets matched. This
    # is because the model argument gets matched with a model subclass that is
    # parametrized with a type that contains the 'Any' below and then mypy runs in
    # to a (false?) ambiguity, due to 'Any', and can't decide what overload/return
    # type to select
    #
    # Example:
    #   class MyModel(models.Model):
    #       field = models.CharField()
    #
    #   # Our plugin auto generates the following subclass
    #   class MyModel_WithAnnotations(MyModel, django_stubs_ext.Annotations[_Annotations]):
    #       ...
    #   # Assume
    #   x = MyModel.objects.annotate(foo=F("id")).get()
    #   reveal_type(x)  # MyModel_WithAnnotations[TypedDict({"foo": Any})]
    #   # Then, on an attribute access of 'field' like
    #   reveal_type(x.field)
    #
    # Now 'CharField.__get__', which is overloaded, is passed 'x' as the 'model'
    # argument and mypy consider it ambiguous to decide which overload method to
    # select due to the 'Any' in 'TypedDict({"foo": Any})'. But if we specify the
    # 'Any' as 'TypeOfAny.special_form' mypy doesn't consider the model instance to
    # contain 'Any' and the ambiguity goes away.
    return {name: AnyType(TypeOfAny.special_form) for name, _ in kwargs.items()}


def extract_proper_type_queryset_annotate(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    django_model = helpers.get_model_info_from_qs_ctx(ctx, django_context)
    if django_model is None:
        return AnyType(TypeOfAny.from_omitted_generics)

    default_return_type = get_proper_type(ctx.default_return_type)
    if not isinstance(default_return_type, Instance):
        return ctx.default_return_type

    api = helpers.get_typechecker_api(ctx)

    expression_types = {
        attr_name: typ
        for attr_name, typ in gather_expression_types(ctx).items()
        if check_valid_attr_value(ctx, django_context, django_model, attr_name)
    }

    annotated_type: ProperType = django_model.typ
    if expression_types:
        fields_dict = helpers.make_typeddict(
            api,
            fields=expression_types,
            required_keys=set(expression_types.keys()),
            readonly_keys=set(),
        )
        annotated_type = get_annotated_type(api, django_model.typ, fields_dict=fields_dict)

    row_type: MypyType
    if len(default_return_type.args) > 1:
        original_row_type = get_proper_type(default_return_type.args[1])
        row_type = original_row_type
        if isinstance(original_row_type, TypedDictType):
            row_type = api.named_generic_type(
                "builtins.dict", [api.named_generic_type("builtins.str", []), AnyType(TypeOfAny.from_omitted_generics)]
            )
        elif isinstance(original_row_type, TupleType):
            if original_row_type.partial_fallback.type.has_base("typing.NamedTuple"):
                # TODO: Use a NamedTuple which contains the known fields, but also
                #  falls back to allowing any attribute access.
                row_type = AnyType(TypeOfAny.implementation_artifact)
            else:
                row_type = api.named_generic_type("builtins.tuple", [AnyType(TypeOfAny.from_omitted_generics)])
        elif isinstance(original_row_type, Instance) and helpers.is_model_type(original_row_type.type):
            row_type = annotated_type
    else:
        row_type = annotated_type
    return default_return_type.copy_modified(args=[annotated_type, row_type])


def resolve_field_lookups(lookup_exprs: Sequence[Expression], django_context: DjangoContext) -> list[str] | None:
    field_lookups = []
    for field_lookup_expr in lookup_exprs:
        field_lookup = helpers.resolve_string_attribute_value(field_lookup_expr, django_context)
        if field_lookup is None:
            return None
        field_lookups.append(field_lookup)
    return field_lookups


def extract_proper_type_queryset_values(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    """
    Extract proper return type for QuerySet.values(*fields, **expressions) method calls.

    See https://docs.djangoproject.com/en/stable/ref/models/querysets/#values
    """
    django_model = helpers.get_model_info_from_qs_ctx(ctx, django_context)
    if django_model is None or django_model.is_annotated:
        return ctx.default_return_type

    default_return_type = get_proper_type(ctx.default_return_type)
    if not isinstance(default_return_type, Instance):
        return ctx.default_return_type

    field_lookups = resolve_field_lookups(ctx.args[0], django_context)
    if field_lookups is None:
        return AnyType(TypeOfAny.from_error)

    # Bare `.values()` case
    if len(field_lookups) == 0 and not ctx.args[1]:
        for field in django_context.get_model_fields(django_model.cls):
            field_lookups.append(field.attname)

    column_types: dict[str, MypyType] = {}

    # Collect `*fields` types -- `.values("id", "name")`
    for field_lookup in field_lookups:
        field_lookup_type = get_field_type_from_lookup(
            ctx, django_context, django_model.cls, lookup=field_lookup, method="values"
        )
        if field_lookup_type is None:
            return default_return_type.copy_modified(args=[django_model.typ, AnyType(TypeOfAny.from_error)])

        column_types[field_lookup] = field_lookup_type

    # Collect `**expressions` types -- `.values(lower_name=Lower("name"), foo=F("name"))`
    column_types.update(gather_expression_types(ctx))
    row_type = helpers.make_typeddict(ctx.api, column_types, set(column_types.keys()), set())
    return default_return_type.copy_modified(args=[django_model.typ, row_type])


def _infer_prefetch_queryset_type(queryset_expr: Expression, api: TypeChecker) -> Instance | None:
    """Infer the model Instance from `Prefetch(queryset=...)`"""
    try:
        qs_type = get_proper_type(api.expr_checker.accept(queryset_expr))
    except Exception:
        return None
    if isinstance(qs_type, Instance):
        return qs_type
    return None


def _resolve_prefetch_string_argument(
    type_arg: MypyType,
    expr: Expression | None,
    django_context: DjangoContext,
    arg_name: str,
) -> str | None:
    # First try to get value from specialized type arg
    arg_value = helpers.get_literal_str_type(type_arg)
    if arg_value is not None:
        return arg_value

    # Fallback: parse inline call expression
    if isinstance(expr, CallExpr):
        arg_expr = helpers.get_class_init_argument_by_name(expr, arg_name)
        if arg_expr:
            return helpers.resolve_string_attribute_value(arg_expr, django_context)
    return None


def _resolve_prefetch_queryset_argument(
    type_arg: MypyType,
    expr: Expression | None,
    api: TypeChecker,
) -> Instance | None:
    # First try to get queryset type from specialized type arg
    queryset_type = get_proper_type(type_arg)
    if isinstance(queryset_type, Instance):
        elem_model = helpers.extract_model_type_from_queryset(queryset_type, api)
        # If we got a valid specific model type, return the queryset type
        if elem_model is not None and elem_model.type.fullname != fullnames.MODEL_CLASS_FULLNAME:
            return queryset_type

    # Fallback: parse inline call expression
    if isinstance(expr, CallExpr):
        queryset_expr = helpers.get_class_init_argument_by_name(expr, "queryset")
        if queryset_expr is not None:
            return _infer_prefetch_queryset_type(queryset_expr, api)

    return None


def _specialize_string_arg_to_literal(
    ctx: FunctionContext, django_context: DjangoContext, arg_name: str, fallback_type: MypyType
) -> MypyType:
    """
    Helper to specialize a string argument to a Literal[str] type.

    This allows the plugin to extract the actual string value for further processing
    and validation in later analysis phases.
    """

    if arg_expr := helpers.get_call_argument_by_name(ctx, arg_name):
        if arg_value := helpers.resolve_string_attribute_value(arg_expr, django_context):
            api = helpers.get_typechecker_api(ctx)
            return LiteralType(value=arg_value, fallback=api.named_generic_type("builtins.str", []))

    return fallback_type


def specialize_prefetch_type(ctx: FunctionContext, django_context: DjangoContext) -> MypyType:
    """Function hook for `Prefetch(...)` to specialize its `lookup` and `to_attr` generic parameters."""
    default = get_proper_type(ctx.default_return_type)
    if not isinstance(default, Instance):
        return ctx.default_return_type

    lookup_type = _specialize_string_arg_to_literal(ctx, django_context, "lookup", default.args[0])
    to_attr_type = _specialize_string_arg_to_literal(ctx, django_context, "to_attr", default.args[2])

    return default.copy_modified(args=[lookup_type, default.args[1], to_attr_type])


def gather_flat_args(ctx: MethodContext) -> list[tuple[Expression | None, ProperType]]:
    """
    Flatten all arguments into a uniform list of (expr, typ) pairs.

    This helper iterates over positional and named arguments and expands any starred
    arguments when their type is a TupleType with statically known items.
    """
    lookups: list[tuple[Expression | None, ProperType]] = []
    arg_start_idx = 0
    for expr, typ, kind in zip(ctx.args[0], ctx.arg_types[0], ctx.arg_kinds[0], strict=False):
        ptyp = get_proper_type(typ)
        if kind == ARG_STAR:
            # Expand starred tuple items if statically known
            if isinstance(ptyp, TupleType):
                lookups.append((None, get_proper_type(ptyp.items[arg_start_idx])))
            # If not a TupleType (e.g. list/Iterable), we cannot expand statically
            arg_start_idx += 1
            continue
        lookups.append((expr, ptyp))
    return lookups


def _get_selected_fields_from_queryset_type(qs_type: Instance) -> set[str] | None:
    """
    Derive selected field names from a QuerySet type.

    Sources:
      - values(): encoded in the row TypedDict keys
      - values_list(named=True): row is a NamedTuple; extract field names from fallback TypeInfo
      - values_list(named=False): stored in qs_type.extra_attrs.immutable
    """
    if len(qs_type.args) > 1:
        row_type = get_proper_type(qs_type.args[1])
        if isinstance(row_type, Instance) and helpers.is_model_type(row_type.type):
            return None
        if isinstance(row_type, TypedDictType):
            return set(row_type.items.keys())
        if isinstance(row_type, TupleType):
            if row_type.partial_fallback.type.has_base("typing.NamedTuple"):
                return {name for name, sym in row_type.partial_fallback.type.names.items() if sym.plugin_generated}
            return set()
        return set()

    # Fallback to explicit metadata attached to the QuerySet Instance
    if qs_type.extra_attrs and qs_type.extra_attrs.immutable and isinstance(qs_type.extra_attrs.immutable, set):
        return qs_type.extra_attrs.immutable

    return None


def check_valid_attr_value(
    ctx: MethodContext,
    django_context: DjangoContext,
    model: DjangoModel,
    attr_name: str,
    *,
    new_attr_names: set[str] | None = None,
) -> bool:
    """
    Check if adding `attr_name` would conflict with existing symbols on `model`.

    Args:
        - model: The Django model being analyzed
        - attr_name: The name of the attribute to be added
        - new_attr_names: A mapping of field names to types currently being added to the model
    """
    deselected_fields: set[str] | None = None
    if isinstance(ctx.type, Instance):
        selected_fields = _get_selected_fields_from_queryset_type(ctx.type)
        if selected_fields is not None:
            model_field_names = {f.name for f in django_context.get_model_fields(model.cls)}
            deselected_fields = model_field_names - selected_fields
            new_attr_names = new_attr_names or set()
            new_attr_names.update(selected_fields - model_field_names)

    is_conflicting_attr_value = bool(
        # 1. Conflict with another symbol on the model (If not de-selected via a prior .values/.values_list call).
        # Ex:
        #     User.objects.prefetch_related(Prefetch(..., to_attr="id"))
        (model.typ.type.get(attr_name) and (deselected_fields is None or attr_name not in deselected_fields))
        # 2. Conflict with a previous annotation.
        # Ex:
        #     User.objects.annotate(foo=...).prefetch_related(Prefetch(...,to_attr="foo"))
        #     User.objects.prefetch_related(Prefetch(...,to_attr="foo")).prefetch_related(Prefetch(...,to_attr="foo"))
        or (model.typ.extra_attrs and attr_name in model.typ.extra_attrs.attrs)
        # 3. Conflict with another symbol added in the current processing.
        # Ex:
        #     User.objects.prefetch_related(
        #        Prefetch("groups", Group.objects.filter(name="test"), to_attr="new_attr"),
        #        Prefetch("groups", Group.objects.all(), to_attr="new_attr"), # E: Not OK!
        #     )
        or (new_attr_names is not None and attr_name in new_attr_names)
    )
    if is_conflicting_attr_value:
        ctx.api.fail(
            f'Attribute "{attr_name}" already defined on "{model.typ}"',
            ctx.context,
            code=NO_REDEF,
        )
    return not is_conflicting_attr_value


def check_valid_prefetch_related_lookup(
    ctx: MethodContext,
    lookup: str,
    django_model: DjangoModel,
    django_context: DjangoContext,
    *,
    is_generic_prefetch: bool = False,
) -> bool:
    """Check if a lookup string resolve to something that can be prefetched"""
    current_model_cls = django_model.cls
    contenttypes_installed = django_context.apps_registry.is_installed("django.contrib.contenttypes")
    for through_attr in lookup.split(LOOKUP_SEP):
        rel_obj_descriptor = getattr(current_model_cls, through_attr, None)
        if rel_obj_descriptor is None:
            ctx.api.fail(
                (
                    f'Cannot find "{through_attr}" on "{current_model_cls.__name__}" object, '
                    f'"{lookup}" is an invalid parameter to "prefetch_related()"'
                ),
                ctx.context,
            )
            return False
        if contenttypes_installed and is_generic_prefetch:
            from django.contrib.contenttypes.fields import GenericForeignKey

            if not isinstance(rel_obj_descriptor, GenericForeignKey):
                ctx.api.fail(
                    f'"{through_attr}" on "{current_model_cls.__name__}" is not a GenericForeignKey, '
                    f"GenericPrefetch can only be used with GenericForeignKey fields",
                    ctx.context,
                )
                return True
        elif isinstance(rel_obj_descriptor, ForwardManyToOneDescriptor):
            current_model_cls = rel_obj_descriptor.field.remote_field.model
        elif isinstance(rel_obj_descriptor, ReverseOneToOneDescriptor):
            current_model_cls = rel_obj_descriptor.related.related_model  # type:ignore[assignment] # Can't be 'self' for non abstract models
        elif isinstance(rel_obj_descriptor, ManyToManyDescriptor):
            current_model_cls = (
                rel_obj_descriptor.rel.related_model if rel_obj_descriptor.reverse else rel_obj_descriptor.rel.model  # type:ignore[assignment] # Can't be 'self' for non abstract models
            )
        elif isinstance(rel_obj_descriptor, ReverseManyToOneDescriptor):
            if contenttypes_installed:
                from django.contrib.contenttypes.fields import ReverseGenericManyToOneDescriptor

                if isinstance(rel_obj_descriptor, ReverseGenericManyToOneDescriptor):
                    current_model_cls = rel_obj_descriptor.rel.model
                    continue
            current_model_cls = rel_obj_descriptor.rel.related_model  # type:ignore[assignment] # Can't be 'self' for non abstract models
        else:
            if contenttypes_installed:
                from django.contrib.contenttypes.fields import GenericForeignKey

                if isinstance(rel_obj_descriptor, GenericForeignKey):
                    # Generic foreign keys can point to any model, so we use Model as the base type
                    return True
            ctx.api.fail(
                (
                    f'"{lookup}" does not resolve to an item that supports prefetching '
                    '- this is an invalid parameter to "prefetch_related()"'
                ),
                ctx.context,
            )
            return False
    return True


def check_conflicting_lookups(
    ctx: MethodContext,
    observed_attr: str,
    qs_types: dict[str, Instance | None],
    queryset_type: Instance | None,
) -> bool:
    is_conflicting_lookup = bool(observed_attr in qs_types and qs_types[observed_attr] != queryset_type)
    if is_conflicting_lookup:
        ctx.api.fail(
            f'Lookup "{observed_attr}" was already seen with a different queryset',
            ctx.context,
            code=NO_REDEF,
        )
    return is_conflicting_lookup


def extract_prefetch_related_annotations(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    """
    Extract annotated attributes via `prefetch_related(Prefetch(..., to_attr=...))`

    See https://docs.djangoproject.com/en/stable/ref/models/querysets/#prefetch-objects
    """
    api = helpers.get_typechecker_api(ctx)

    if not (
        isinstance(ctx.type, Instance)
        and isinstance((default_return_type := get_proper_type(ctx.default_return_type)), Instance)
        and (qs_model := helpers.get_model_info_from_qs_ctx(ctx, django_context)) is not None
        and ctx.args
        and ctx.arg_types
        and ctx.arg_types[0]
        # Only process the correct overload, i.e.
        #     def prefetch_related(self, *lookups: str | Prefetch[_PrefetchedQuerySetT, _ToAttrT]) -> Self: ...
        and None not in ctx.callee_arg_names
    ):
        return ctx.default_return_type

    new_attrs: dict[str, MypyType] = {}  # A mapping of field_name / types to add to the model
    qs_types: dict[str, Instance | None] = {}  # A mapping of field_name / associated queryset type
    for expr, typ in gather_flat_args(ctx):
        if not (isinstance(typ, Instance) and typ.type.has_base(fullnames.PREFETCH_CLASS_FULLNAME)):
            # Handle plain string lookups (not Prefetch instances)
            lookup = helpers.get_literal_str_type(typ)
            queryset_type = None
            if lookup is not None:
                check_valid_prefetch_related_lookup(ctx, lookup, qs_model, django_context)
                check_conflicting_lookups(ctx, lookup, qs_types, queryset_type)
                qs_types[lookup] = queryset_type
            continue

        # 1) Extract lookup value from specialized type arg or call expression
        lookup = _resolve_prefetch_string_argument(typ.args[0], expr, django_context, "lookup")

        # 2) Extract to_attr value from specialized type arg or call expression
        to_attr = _resolve_prefetch_string_argument(typ.args[2], expr, django_context, "to_attr")
        if to_attr is None and lookup is None:
            continue

        # 3.a) Determine queryset type from specialized type arg or call expression
        queryset_type = _resolve_prefetch_queryset_argument(typ.args[1], expr, api)

        # 3.b) Extract model type from queryset type (or from the lookup value)
        elem_model: Instance | None = None
        if queryset_type is not None and isinstance(queryset_type, Instance):
            elem_model = helpers.extract_model_type_from_queryset(queryset_type, api)
        elif lookup:
            try:
                observed_model_cls = django_context.resolve_lookup_into_field(qs_model.cls, lookup)[1]
                if model_info := helpers.lookup_class_typeinfo(api, observed_model_cls):
                    elem_model = Instance(model_info, [])
            except (FieldError, LookupsAreUnsupported):
                pass

        if to_attr and check_valid_attr_value(
            ctx, django_context, qs_model, to_attr, new_attr_names=set(new_attrs.keys())
        ):
            # When traversing multiple relations (e.g. "groups__user_set"), the to_attr is set
            # on the last item of the chain (e.g. Group), not on the root model (e.g. User).
            # We can't annotate an intermediate model from here, so skip adding the annotation
            # to the root model to avoid incorrectly attributing to_attr to it.
            if lookup and LOOKUP_SEP in lookup:
                continue
            new_attrs[to_attr] = api.named_generic_type(
                "builtins.list",
                [elem_model if elem_model is not None else AnyType(TypeOfAny.special_form)],
            )
            qs_types[to_attr] = queryset_type
        if not to_attr and lookup:
            check_valid_prefetch_related_lookup(
                ctx,
                lookup,
                qs_model,
                django_context,
                is_generic_prefetch=typ.type.has_base(fullnames.GENERIC_PREFETCH_CLASS_FULLNAME),
            )
            check_conflicting_lookups(ctx, lookup, qs_types, queryset_type)
            qs_types[lookup] = queryset_type

    if not new_attrs:
        return ctx.default_return_type

    fields_dict = helpers.make_typeddict(
        api,
        fields=new_attrs,
        required_keys=set(new_attrs.keys()),
        readonly_keys=set(),
    )

    annotated_model = get_annotated_type(api, qs_model.typ, fields_dict=fields_dict)

    # Keep row shape; if row is a model instance, update it to annotated
    # Todo: consolidate with `extract_proper_type_queryset_annotate` row handling above.
    if len(default_return_type.args) > 1:
        original_row = get_proper_type(default_return_type.args[1])
        row_type: MypyType = original_row
        if isinstance(original_row, Instance) and helpers.is_model_type(original_row.type):
            row_type = annotated_model
    else:
        row_type = annotated_model

    return default_return_type.copy_modified(args=[annotated_model, row_type])


def _try_get_field(
    ctx: MethodContext, model_cls: type[Model], field_name: str, *, resolve_pk: bool = False
) -> "_AnyField | None":
    opts = model_cls._meta
    resolved_name = opts.pk.name if resolve_pk and field_name == "pk" else field_name
    try:
        return opts.get_field(resolved_name)
    except FieldDoesNotExist as e:
        ctx.api.fail(str(e), ctx.context)
        return None


def _check_field_concrete(ctx: MethodContext, field: "_AnyField", field_name: str, method: str) -> bool:
    if not field.concrete or field.many_to_many:
        ctx.api.fail(f'"{method}()" can only be used with concrete fields. Got "{field_name}"', ctx.context)
        return False
    return True


def _check_field_not_pk(
    ctx: MethodContext,
    model_cls: type[Model],
    field: "_AnyField",
    field_name: str,
    method: str,
    *,
    attr_name: str | None = None,
) -> bool:
    opts = model_cls._meta
    all_pk_fields = set(getattr(opts, "pk_fields", [opts.pk]))
    for parent in getattr(opts, "all_parents", opts.get_parent_list()):
        all_pk_fields.update(getattr(parent._meta, "pk_fields", [parent._meta.pk]))
    if field in all_pk_fields:
        param_str = f' in "{attr_name}="' if attr_name else ""
        ctx.api.fail(f'"{method}()" does not support primary key fields{param_str}. Got "{field_name}"', ctx.context)
        return False
    return True


def _extract_field_names_from_collection(
    ctx: MethodContext, django_context: DjangoContext, arg_index: int
) -> list[str] | None:
    if not (
        len(ctx.args) > arg_index
        and ctx.args[arg_index]
        and isinstance((collection_expr := ctx.args[arg_index][0]), (ListExpr, TupleExpr, SetExpr))
    ):
        return None

    return [
        field_name
        for field_arg in collection_expr.items
        if (field_name := helpers.resolve_string_attribute_value(field_arg, django_context)) is not None
    ]


def _extract_field_names_from_varargs(ctx: MethodContext) -> list[str]:
    return [
        lookup_value
        for lookup_type in (ctx.arg_types[0] if ctx.arg_types and ctx.arg_types[0] else [])
        if (lookup_value := helpers.get_literal_str_type(get_proper_type(lookup_type))) is not None
    ]


def _get_select_related_field_choices(model_cls: type[Model]) -> set[str]:
    """
    Get valid field choices for select_related lookups.
    Based on Django's SQLCompiler.get_related_selections._get_field_choices method.
    """
    opts = model_cls._meta

    # Direct relation fields (forward relations)
    direct_choices = (f.name for f in opts.fields if f.is_relation)

    # Reverse relation fields (backward relations with unique=True)
    reverse_choices = (f.field.related_query_name() for f in opts.related_objects if f.field.unique)
    return {*direct_choices, *reverse_choices}


def _validate_select_related_lookup(
    ctx: MethodContext,
    django_context: DjangoContext,
    model_cls: type[Model],
    lookup: str,
) -> bool:
    """Validate a single select_related lookup string."""
    if not lookup.strip():
        ctx.api.fail(
            f'Invalid field name "{lookup}" in select_related lookup',
            ctx.context,
        )
        return False

    lookup_parts = lookup.split(LOOKUP_SEP)
    observed_model = model_cls
    for i, part in enumerate(lookup_parts):
        valid_choices = _get_select_related_field_choices(observed_model)

        if part not in valid_choices:
            ctx.api.fail(
                f'Invalid field name "{part}" in select_related lookup. '
                f"Choices are: {', '.join(sorted(valid_choices)) or '(none)'}",
                ctx.context,
            )
            return False

        if i < len(lookup_parts) - 1:  # Not the last part
            try:
                field, observed_model = django_context.resolve_lookup_into_field(observed_model, part)
                if field is None:
                    return False
            except (FieldError, LookupsAreUnsupported):
                # For good measure, but we should never reach this since we already validated the part name
                return False

    return True


def validate_select_related(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    """
    Validates that all lookup strings passed to select_related() resolve to actual model fields and relations.

    Extracted and adapted from `django.db.models.sql.compiler.SQLCompiler.get_related_selections`
    """
    if (django_model := helpers.get_model_info_from_qs_ctx(ctx, django_context)) is None:
        return ctx.default_return_type

    for lookup_value in _extract_field_names_from_varargs(ctx):
        _validate_select_related_lookup(ctx, django_context, django_model.cls, lookup_value)

    return ctx.default_return_type


def _validate_bulk_update_field(
    ctx: MethodContext, model_cls: type[Model], field_name: str, method: str, *, attr_name: str | None = None
) -> bool:
    return (
        (field := _try_get_field(ctx, model_cls, field_name)) is not None
        and _check_field_concrete(ctx, field, field_name, method)
        and _check_field_not_pk(ctx, model_cls, field, field_name, method, attr_name=attr_name)
    )


def validate_bulk_update(
    ctx: MethodContext, django_context: DjangoContext, method: Literal["bulk_update", "abulk_update"]
) -> MypyType:
    """
    Type check the `fields` argument passed to `QuerySet.bulk_update(...)`.

    Extracted and adapted from `django.db.models.query.QuerySet.bulk_update`
    Mirrors tests from `django/tests/queries/test_bulk_update.py`
    """
    if (django_model := helpers.get_model_info_from_qs_ctx(ctx, django_context)) is None or (
        field_names := _extract_field_names_from_collection(ctx, django_context, arg_index=1)
    ) is None:
        return ctx.default_return_type

    if not field_names:
        # Only error if the collection literal itself is empty (not when items are unresolvable).
        collection_expr = ctx.args[1][0]
        if isinstance(collection_expr, (ListExpr, TupleExpr, SetExpr)) and not collection_expr.items:
            ctx.api.fail(f'Field names must be given to "{method}()"', ctx.context)
        return ctx.default_return_type

    for field_name in field_names:
        _validate_bulk_update_field(ctx, django_model.cls, field_name, method)

    return ctx.default_return_type


def _validate_bulk_create_unique_field(
    ctx: MethodContext, model_cls: type[Model], field_name: str, method: str
) -> bool:
    return (field := _try_get_field(ctx, model_cls, field_name, resolve_pk=True)) is not None and _check_field_concrete(
        ctx, field, field_name, method
    )


def validate_bulk_create(
    ctx: MethodContext, django_context: DjangoContext, method: Literal["bulk_create", "abulk_create"]
) -> MypyType:
    """
    Type check the `update_fields` and `unique_fields` arguments passed to `QuerySet.bulk_create(...)`.

    Extracted and adapted from `django.db.models.query.QuerySet._check_bulk_create_options`
    """
    if (django_model := helpers.get_model_info_from_qs_ctx(ctx, django_context)) is None:
        return ctx.default_return_type

    update_field_names = _extract_field_names_from_collection(ctx, django_context, arg_index=4)
    if update_field_names is not None:
        for field_name in update_field_names:
            _validate_bulk_update_field(ctx, django_model.cls, field_name, method, attr_name="update_fields")

    unique_field_names = _extract_field_names_from_collection(ctx, django_context, arg_index=5)
    if unique_field_names is not None:
        for field_name in unique_field_names:
            _validate_bulk_create_unique_field(ctx, django_model.cls, field_name, method)

    return ctx.default_return_type


def _validate_order_by_lookup(ctx: MethodContext, model_cls: type[Model], parts: list[str]) -> None:
    if len(parts) == 1 and parts[0] == "?":
        return

    # Abstract models don't have a pk field, skip validation
    if model_cls._meta.abstract:
        return

    try:
        _, final_field, _, remainder = Query(model_cls).names_to_path(parts, model_cls._meta)
    except FieldError as exc:
        ctx.api.fail(exc.args[0], ctx.context)
        return

    if remainder:
        # Check if the trailing part is a valid transform (e.g. __year, __month) on the field.
        # Transforms are allowed in order_by, but lookups (e.g. __exact) are not.
        lookup_cls = final_field.get_lookups().get(remainder[0])
        if lookup_cls is None or not issubclass(lookup_cls, Transform):
            msg = f"Cannot resolve keyword '{remainder[0]}' into field or transform on '{final_field.name}'."
            ctx.api.fail(msg, ctx.context)


def validate_order_by(ctx: MethodContext, django_context: DjangoContext) -> MypyType:
    if (django_model := helpers.get_model_info_from_qs_ctx(ctx, django_context)) is None:
        return ctx.default_return_type

    selected_fields = _get_selected_fields_from_queryset_type(ctx.type) if isinstance(ctx.type, Instance) else None

    for lookup_value in _extract_field_names_from_varargs(ctx):
        parts = lookup_value.removeprefix("-").split(LOOKUP_SEP)

        if django_model.typ.extra_attrs and parts[0] in django_model.typ.extra_attrs.attrs:
            # Skip validation for annotated fields
            continue
        if selected_fields is not None and parts[0] in selected_fields:
            # Skip validation for fields selected via values()/values_list()
            continue
        _validate_order_by_lookup(ctx, django_model.cls, parts)

    return ctx.default_return_type


def _validate_defer_only_fields(
    ctx: MethodContext, model_cls: type[Model], field_names: list[str], *, is_defer: bool
) -> None:
    query = Query(model_cls)
    query.add_deferred_loading(field_names) if is_defer else query.add_immediate_loading(field_names)

    try:
        query.get_select_mask()
    except (FieldDoesNotExist, FieldError) as exc:
        method = "defer" if is_defer else "only"
        ctx.api.fail(f'Invalid field in "{method}()": {exc.args[0]}', ctx.context)


def validate_defer_only(ctx: MethodContext, django_context: DjangoContext, *, is_defer: bool) -> MypyType:
    if (django_model := helpers.get_model_info_from_qs_ctx(ctx, django_context)) is None or not (
        field_names := _extract_field_names_from_varargs(ctx)
    ):
        return ctx.default_return_type

    _validate_defer_only_fields(ctx, django_model.cls, field_names, is_defer=is_defer)

    return ctx.default_return_type



import itertools
import sys
from collections.abc import Callable
from functools import cached_property, partial
from typing import Any

from mypy.build import PRI_MED, PRI_MYPY
from mypy.modulefinder import mypy_path
from mypy.nodes import MypyFile, TypeInfo
from mypy.options import Options
from mypy.plugin import (
    AnalyzeTypeContext,
    AttributeContext,
    ClassDefContext,
    DynamicClassDefContext,
    FunctionContext,
    MethodContext,
    Plugin,
    ReportConfigContext,
)
from mypy.types import Type as MypyType
from typing_extensions import override

from mypy_django_plugin.config import DjangoPluginConfig
from mypy_django_plugin.django.context import DjangoContext
from mypy_django_plugin.exceptions import UnregisteredModelError
from mypy_django_plugin.lib import fullnames, helpers
from mypy_django_plugin.transformers import (
    choices,
    fields,
    forms,
    init_create,
    manytomany,
    manytoone,
    meta,
    orm_lookups,
    querysets,
    settings,
)
from mypy_django_plugin.transformers.auth import get_user_model
from mypy_django_plugin.transformers.functional import resolve_str_promise_attribute
from mypy_django_plugin.transformers.managers import (
    add_as_manager_to_queryset_class,
    create_new_manager_class_from_from_queryset_method,
    reparametrize_any_manager_hook,
    resolve_manager_method,
)
from mypy_django_plugin.transformers.models import (
    MetaclassAdjustments,
    handle_annotated_type,
    process_model_class,
    set_auth_user_model_boolean_fields,
)
from mypy_django_plugin.transformers.request import check_querydict_is_mutable


class NewSemanalDjangoPlugin(Plugin):
    def __init__(self, options: Options) -> None:
        super().__init__(options)
        self.plugin_config = DjangoPluginConfig(options.config_file)
        # Add paths from MYPYPATH env var
        sys.path.extend(mypy_path())
        # Add paths from mypy_path config option
        sys.path.extend(options.mypy_path)
        self.django_context = DjangoContext(self.plugin_config.django_settings_module)

    def _get_current_form_bases(self) -> dict[str, int]:
        model_info = self._get_typeinfo_or_none(fullnames.BASEFORM_CLASS_FULLNAME)
        if model_info:
            bases = helpers.get_django_metadata_bases(model_info, "baseform_bases")
            bases[fullnames.BASEFORM_CLASS_FULLNAME] = 1
            bases[fullnames.FORM_CLASS_FULLNAME] = 1
            bases[fullnames.MODELFORM_CLASS_FULLNAME] = 1
            return bases
        return {}

    def _get_typeinfo_or_none(self, class_name: str) -> TypeInfo | None:
        sym = self.lookup_fully_qualified(class_name)
        if sym is not None and isinstance(sym.node, TypeInfo):
            return sym.node
        return None

    def _new_dependency(self, module: str, priority: int = PRI_MYPY) -> tuple[int, str, int]:
        fake_lineno = -1
        return (priority, module, fake_lineno)

    @override
    def get_additional_deps(self, file: MypyFile) -> list[tuple[int, str, int]]:
        # for settings
        if file.fullname == "django.conf" and self.django_context.django_settings_module:
            return [self._new_dependency(self.django_context.django_settings_module, PRI_MED)]

        # for values / values_list
        if file.fullname == "django.db.models":
            return [self._new_dependency("typing"), self._new_dependency("django_stubs_ext")]

        # for `get_user_model()`
        if file.fullname == "django.contrib.auth" or file.fullname in {"django.http", "django.http.request"}:
            auth_user_model_name = self.django_context.settings.AUTH_USER_MODEL
            try:
                auth_user_module = self.django_context.apps_registry.get_model(auth_user_model_name).__module__
            except LookupError:
                # get_user_model() model app is not installed
                return []
            return [self._new_dependency(auth_user_module), self._new_dependency("django_stubs_ext")]

        # ensure that all mentions to='someapp.SomeModel' are loaded with corresponding related Fields
        defined_model_classes = self.django_context.model_modules.get(file.fullname)
        if not defined_model_classes:
            return []
        deps = set()

        for model_class in defined_model_classes.values():
            for field in itertools.chain(
                # forward relations
                self.django_context.get_model_related_fields(model_class),
                # reverse relations - `related_objects` is private API (according to docstring)
                model_class._meta.related_objects,
            ):
                try:
                    related_model_cls = self.django_context.get_field_related_model_cls(field)  # type: ignore[arg-type]
                except UnregisteredModelError:
                    continue
                related_model_module = related_model_cls.__module__
                if related_model_module != file.fullname:
                    deps.add(self._new_dependency(related_model_module))

        return [
            *deps,
            # For `QuerySet.annotate`
            self._new_dependency("django_stubs_ext"),
            # For `TypedModelMeta` lookup in model transformers
            self._new_dependency("django_stubs_ext.db.models"),
            # For `Manager.from_queryset`
            self._new_dependency("django.db.models.query"),
        ]

    @override
    def get_function_hook(self, fullname: str) -> Callable[[FunctionContext], MypyType] | None:
        info = self._get_typeinfo_or_none(fullname)
        if info:
            if info.has_base(fullnames.FIELD_FULLNAME):
                return partial(fields.transform_into_proper_return_type, django_context=self.django_context)

            if helpers.is_model_type(info):
                return partial(init_create.typecheck_model_init, django_context=self.django_context)

            if info.has_base(fullnames.BASE_MANAGER_CLASS_FULLNAME):
                return querysets.determine_proper_manager_type

            if info.has_base(fullnames.PREFETCH_CLASS_FULLNAME):
                return partial(querysets.specialize_prefetch_type, django_context=self.django_context)

        return None

    @cached_property
    def manager_and_queryset_method_hooks(self) -> dict[str, Callable[[MethodContext], MypyType]]:
        typecheck_filtering_method = partial(orm_lookups.typecheck_queryset_filter, django_context=self.django_context)
        return {
            "values": partial(querysets.extract_proper_type_queryset_values, django_context=self.django_context),
            "values_list": partial(
                querysets.extract_proper_type_queryset_values_list, django_context=self.django_context
            ),
            "alias": partial(querysets.extract_proper_type_queryset_annotate, django_context=self.django_context),
            "annotate": partial(querysets.extract_proper_type_queryset_annotate, django_context=self.django_context),
            "create": partial(init_create.typecheck_model_create, django_context=self.django_context),
            "acreate": partial(init_create.typecheck_model_acreate, django_context=self.django_context),
            "filter": typecheck_filtering_method,
            "get": typecheck_filtering_method,
            "aget": typecheck_filtering_method,
            "exclude": typecheck_filtering_method,
            "prefetch_related": partial(
                querysets.extract_prefetch_related_annotations, django_context=self.django_context
            ),
            "select_related": partial(querysets.validate_select_related, django_context=self.django_context),
            "bulk_update": partial(
                querysets.validate_bulk_update, django_context=self.django_context, method="bulk_update"
            ),
            "abulk_update": partial(
                querysets.validate_bulk_update, django_context=self.django_context, method="abulk_update"
            ),
            "bulk_create": partial(
                querysets.validate_bulk_create, django_context=self.django_context, method="bulk_create"
            ),
            "abulk_create": partial(
                querysets.validate_bulk_create, django_context=self.django_context, method="abulk_create"
            ),
            "order_by": partial(querysets.validate_order_by, django_context=self.django_context),
            "earliest": partial(querysets.validate_order_by, django_context=self.django_context),
            "aearliest": partial(querysets.validate_order_by, django_context=self.django_context),
            "latest": partial(querysets.validate_order_by, django_context=self.django_context),
            "alatest": partial(querysets.validate_order_by, django_context=self.django_context),
            "defer": partial(querysets.validate_defer_only, django_context=self.django_context, is_defer=True),
            "only": partial(querysets.validate_defer_only, django_context=self.django_context, is_defer=False),
        }

    @override
    def get_method_hook(self, fullname: str) -> Callable[[MethodContext], MypyType] | None:
        class_fullname, _, method_name = fullname.rpartition(".")
        # Methods called very often -- short circuit for minor speed up
        if method_name == "__init_subclass__" or fullname.startswith("builtins."):
            return None

        if class_fullname.endswith("QueryDict"):
            info = self._get_typeinfo_or_none(class_fullname)
            if info and info.has_base(fullnames.QUERYDICT_CLASS_FULLNAME):
                return check_querydict_is_mutable

        elif method_name == "__get__":
            hooks = {
                fullnames.MANYTOMANY_FIELD_FULLNAME: manytomany.refine_many_to_many_related_manager,
                fullnames.MANY_TO_MANY_DESCRIPTOR: manytomany.refine_many_to_many_related_manager,
                fullnames.REVERSE_MANY_TO_ONE_DESCRIPTOR: manytoone.refine_many_to_one_related_manager,
            }
            return hooks.get(class_fullname)

        if method_name in self.manager_and_queryset_method_hooks:
            info = self._get_typeinfo_or_none(class_fullname)
            if info and helpers.has_any_of_bases(
                info, [fullnames.QUERYSET_CLASS_FULLNAME, fullnames.MANAGER_CLASS_FULLNAME]
            ):
                return self.manager_and_queryset_method_hooks[method_name]
        elif method_name == "get_field":
            info = self._get_typeinfo_or_none(class_fullname)
            if info and info.has_base(fullnames.OPTIONS_CLASS_FULLNAME):
                return partial(meta.return_proper_field_type_from_get_field, django_context=self.django_context)

        return None

    @override
    def get_customize_class_mro_hook(self, fullname: str) -> Callable[[ClassDefContext], None] | None:
        info = self._get_typeinfo_or_none(fullname)
        if info and info.has_base(fullnames.BASE_MANAGER_CLASS_FULLNAME):
            return reparametrize_any_manager_hook
        return None

    @override
    def get_metaclass_hook(self, fullname: str) -> Callable[[ClassDefContext], None] | None:
        if fullname == fullnames.MODEL_METACLASS_FULLNAME:
            return partial(MetaclassAdjustments.adjust_model_class, plugin_config=self.plugin_config)
        return None

    @override
    def get_base_class_hook(self, fullname: str) -> Callable[[ClassDefContext], None] | None:
        # Base class is a Model class definition
        info = self._get_typeinfo_or_none(fullname)
        if info and helpers.is_model_type(info):
            return partial(process_model_class, django_context=self.django_context)

        # Base class is a Form class definition
        if fullname in self._get_current_form_bases():
            return forms.transform_form_class

        # Base class is a QuerySet class definition
        if info and info.has_base(fullnames.QUERYSET_CLASS_FULLNAME):
            return add_as_manager_to_queryset_class
        return None

    @override
    def get_attribute_hook(self, fullname: str) -> Callable[[AttributeContext], MypyType] | None:
        class_name, _, attr_name = fullname.rpartition(".")

        # Lookup of a settings variable
        if class_name == fullnames.DUMMY_SETTINGS_BASE_CLASS:
            return partial(
                settings.get_type_of_settings_attribute,
                django_context=self.django_context,
                plugin_config=self.plugin_config,
            )

        info = self._get_typeinfo_or_none(class_name)

        # Lookup of the '.is_superuser' attribute
        if info and info.has_base(fullnames.PERMISSION_MIXIN_CLASS_FULLNAME) and attr_name == "is_superuser":
            return partial(set_auth_user_model_boolean_fields, django_context=self.django_context)

        # Lookup of the 'user.is_staff' or 'user.is_active' attribute
        if info and info.has_base(fullnames.ABSTRACT_USER_MODEL_FULLNAME) and attr_name in ("is_staff", "is_active"):
            return partial(set_auth_user_model_boolean_fields, django_context=self.django_context)

        # Lookup of a method on a dynamically generated manager class
        # i.e. a manager class only existing while mypy is running, not collected from the AST
        if (
            info
            and info.has_base(fullnames.BASE_MANAGER_CLASS_FULLNAME)
            and "from_queryset_manager" in helpers.get_django_metadata(info)
        ):
            return resolve_manager_method

        if info and info.has_base(fullnames.STR_PROMISE_FULLNAME):
            return resolve_str_promise_attribute

        if info and (
            (
                info.has_base(fullnames.CHOICES_TYPE_METACLASS_FULLNAME)
                and attr_name in {"choices", "labels", "values", "__empty__"}
            )
            or (info.has_base(fullnames.CHOICES_CLASS_FULLNAME) and attr_name in {"label", "value"})
        ):
            return choices.transform_into_proper_attr_type

        return None

    @override
    def get_type_analyze_hook(self, fullname: str) -> Callable[[AnalyzeTypeContext], MypyType] | None:
        if fullname in fullnames.ANNOTATED_TYPES_FULLNAMES:
            return partial(handle_annotated_type, fullname=fullname)
        if fullname == "django.contrib.auth.models._User":
            return partial(get_user_model, django_context=self.django_context)
        return None

    @override
    def get_dynamic_class_hook(self, fullname: str) -> Callable[[DynamicClassDefContext], None] | None:
        # Create a new manager class definition when a manager's '.from_queryset' classmethod is called
        class_name, _, method_name = fullname.rpartition(".")
        if method_name == "from_queryset":
            info = self._get_typeinfo_or_none(class_name)
            if info and info.has_base(fullnames.BASE_MANAGER_CLASS_FULLNAME):
                return create_new_manager_class_from_from_queryset_method
        return None

    @override
    def report_config_data(self, ctx: ReportConfigContext) -> dict[str, Any]:
        # Cache would be cleared if any settings do change.
        extra_data = {
            "AUTH_USER_MODEL": self.django_context.settings.AUTH_USER_MODEL,
        }
        return self.plugin_config.to_json(extra_data)


def plugin(version: str) -> type[NewSemanalDjangoPlugin]:
    return NewSemanalDjangoPlugin



import logging
import sys
from argparse import ArgumentParser
from collections.abc import Callable, Iterable, Iterator, Sequence
from contextlib import AbstractContextManager
from typing import Any, Literal
from unittest import TestCase, TestLoader, TestResult, TestSuite, TextTestResult, TextTestRunner

from django.db.backends.base.base import BaseDatabaseWrapper
from django.test.testcases import SimpleTestCase
from django.test.testcases import TestCase as DjangoTestCase
from django.test.utils import TimeKeeperProtocol
from typing_extensions import override

class QueryFormatter(logging.Formatter): ...

class DebugSQLTextTestResult(TextTestResult):
    logger: logging.Logger
    def __init__(self, stream: Any, descriptions: bool, verbosity: int) -> None: ...
    handler: logging.StreamHandler[Any]
    @override
    def startTest(self, test: TestCase) -> None: ...
    @override
    def stopTest(self, test: TestCase) -> None: ...
    @override
    def addError(self, test: Any, err: Any) -> None: ...
    @override
    def addFailure(self, test: Any, err: Any) -> None: ...

class PDBDebugResult(TextTestResult):
    def debug(self, error: tuple[type[BaseException], BaseException, Any]) -> None: ...

class DummyList:
    __slots__ = ()
    def append(self, item: Any) -> None: ...

class RemoteTestResult(TestResult):
    events: list[Any]
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    @property
    def test_index(self) -> int: ...
    def _confirm_picklable(self, obj: Any) -> None: ...
    def check_picklable(self, test: Any, err: Any) -> None: ...
    def check_subtest_picklable(self, test: Any, subtest: Any) -> None: ...
    @override
    def startTestRun(self) -> None: ...
    @override
    def stopTestRun(self) -> None: ...
    @override
    def startTest(self, test: Any) -> None: ...
    @override
    def stopTest(self, test: Any) -> None: ...
    if sys.version_info >= (3, 12):
        @override
        def addDuration(self, test: Any, elapsed: Any) -> None: ...
    else:
        def addDuration(self, test: Any, elapsed: Any) -> None: ...
    @override
    def addError(self, test: Any, err: Any) -> None: ...
    @override
    def addFailure(self, test: Any, err: Any) -> None: ...
    @override
    def addSubTest(self, test: Any, subtest: Any, err: Any) -> None: ...
    @override
    def addSuccess(self, test: Any) -> None: ...
    @override
    def addSkip(self, test: Any, reason: Any) -> None: ...
    @override
    def addExpectedFailure(self, test: Any, err: Any) -> None: ...
    @override
    def addUnexpectedSuccess(self, test: Any) -> None: ...
    @override
    def wasSuccessful(self) -> bool: ...

class RemoteTestRunner:
    resultclass: Any
    failfast: bool
    buffer: bool
    def __init__(self, failfast: bool = ..., resultclass: Any | None = ..., buffer: bool = ...) -> None: ...
    def run(self, test: Any) -> Any: ...

def get_max_test_processes() -> int: ...
def parallel_type(value: str) -> int | Literal["auto"]: ...

class ParallelTestSuite(TestSuite):
    init_worker: Callable[..., Any]
    process_setup: Callable[..., Any]
    process_setup_args: tuple[Any, ...]
    run_subsuite: Callable[..., Any]
    runner_class: type[RemoteTestRunner]
    subsuites: list[TestSuite]
    processes: int
    failfast: bool
    debug_mode: bool
    buffer: bool
    initial_settings: dict[str, dict[str, Any]] | None
    serialized_contents: dict[str, str] | None
    used_aliases: set[str] | None
    def __init__(
        self,
        subsuites: list[TestSuite],
        processes: int,
        failfast: bool = ...,
        debug_mode: bool = ...,
        buffer: bool = ...,
    ) -> None: ...
    @override
    def run(self, result: Any) -> Any: ...  # type: ignore[override]
    def handle_event(self, result: Any, tests: list[TestSuite], event: Sequence[Any]) -> None: ...
    def initialize_suite(self) -> None: ...

class Shuffler:
    hash_algorithm: str
    seed: int
    seed_source: str
    def __init__(self, seed: int | None = ...) -> None: ...
    @property
    def seed_display(self) -> str: ...
    def shuffle(self, items: Iterable[Any], key: Callable[[Any], str]) -> list[Any]: ...

class DiscoverRunner:
    test_suite: type[TestSuite]
    parallel_test_suite: type[ParallelTestSuite]
    test_runner: type[TextTestRunner]
    test_loader: TestLoader
    reorder_by: tuple[type[DjangoTestCase], type[SimpleTestCase]]
    pattern: str | None
    top_level: str | None
    verbosity: int
    interactive: bool
    failfast: bool
    keepdb: bool
    reverse: bool
    debug_mode: bool
    debug_sql: bool
    parallel: int
    tags: set[str]
    exclude_tags: set[str]
    pdb: bool
    buffer: bool
    test_name_patterns: set[str] | None
    time_keeper: TimeKeeperProtocol
    shuffle: int | Literal[False]
    logger: logging.Logger | None
    durations: int | None
    def __init__(
        self,
        pattern: str | None = ...,
        top_level: str | None = ...,
        verbosity: int = ...,
        interactive: bool = ...,
        failfast: bool = ...,
        keepdb: bool = ...,
        reverse: bool = ...,
        debug_mode: bool = ...,
        debug_sql: bool = ...,
        parallel: int = ...,
        tags: list[str] | None = ...,
        exclude_tags: list[str] | None = ...,
        test_name_patterns: list[str] | None = ...,
        pdb: bool = ...,
        buffer: bool = ...,
        enable_faulthandler: bool = ...,
        timing: bool = ...,
        shuffle: int | Literal[False] = ...,
        logger: logging.Logger | None = ...,
        durations: int | None = ...,
        **kwargs: Any,
    ) -> None: ...
    @classmethod
    def add_arguments(cls, parser: ArgumentParser) -> None: ...
    @property
    def shuffle_seed(self) -> int | None: ...
    def log(self, msg: str, level: int | None = ...) -> None: ...
    def setup_test_environment(self, **kwargs: Any) -> None: ...
    def setup_shuffler(self) -> None: ...
    def load_with_patterns(self) -> AbstractContextManager[None]: ...
    def load_tests_for_label(self, label: str, discover_kwargs: dict[str, str]) -> TestSuite: ...
    def build_suite(self, test_labels: Sequence[str] | None = ..., **kwargs: Any) -> TestSuite: ...
    def setup_databases(self, **kwargs: Any) -> list[tuple[BaseDatabaseWrapper, str, bool]]: ...
    def get_resultclass(self) -> type[TextTestResult] | None: ...
    def get_test_runner_kwargs(self) -> dict[str, Any]: ...
    def run_checks(self, databases: set[str]) -> None: ...
    def run_suite(self, suite: TestSuite, **kwargs: Any) -> TextTestResult: ...
    def teardown_databases(self, old_config: list[tuple[BaseDatabaseWrapper, str, bool]], **kwargs: Any) -> None: ...
    def teardown_test_environment(self, **kwargs: Any) -> None: ...
    def suite_result(self, suite: TestSuite, result: TextTestResult, **kwargs: Any) -> int: ...
    def _get_databases(self, suite: TestSuite) -> set[str]: ...
    def get_databases(self, suite: TestSuite) -> set[str]: ...
    def run_tests(self, test_labels: list[str], **kwargs: Any) -> int: ...

def try_importing(label: str) -> tuple[bool, bool]: ...
def find_top_level(top_level: str) -> str: ...
def shuffle_tests(tests: Iterable[TestCase], shuffler: Shuffler) -> Iterator[TestCase]: ...
def reorder_test_bin(
    tests: Sequence[TestCase], shuffler: Shuffler | None = ..., reverse: bool = ...
) -> Iterator[TestCase]: ...
def reorder_tests(
    tests: Iterable[TestCase],
    classes: Sequence[type[TestCase]],
    reverse: bool = ...,
    shuffler: Shuffler | None = ...,
) -> Iterator[TestCase]: ...
def partition_suite_by_case(suite: TestSuite) -> list[TestSuite]: ...
def test_match_tags(test: TestCase, tags: set[str], exclude_tags: set[str]) -> bool: ...
def filter_tests_by_tags(tests: Iterable[TestCase], tags: set[str], exclude_tags: set[str]) -> Iterator[TestCase]: ...



from collections.abc import Callable

from django.http.request import HttpRequest
from django.utils.functional import _StrPromise
from django.utils.safestring import SafeString

def csrf_input(request: HttpRequest) -> SafeString: ...

csrf_input_lazy: Callable[[HttpRequest], SafeString]
csrf_token_lazy: Callable[[HttpRequest], _StrPromise]



from collections.abc import Iterable, Sequence

from django.core.checks import CheckMessage
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models import Deferrable
from django.db.models.base import Model
from django.db.models.constraints import BaseConstraint
from django.db.models.expressions import Combinable
from django.db.models.indexes import IndexExpression
from django.db.models.query_utils import Q
from django.utils.functional import _StrOrPromise
from typing_extensions import override

from .utils import CheckPostgresInstalledMixin

class ExclusionConstraintExpression(IndexExpression): ...

class ExclusionConstraint(CheckPostgresInstalledMixin, BaseConstraint):
    template: str
    expressions: Sequence[tuple[str | Combinable, str]]
    index_type: str
    condition: Q | None
    def __init__(
        self,
        *,
        name: str,
        expressions: Sequence[tuple[str | Combinable, str]],
        index_type: str | None = None,
        condition: Q | None = None,
        deferrable: Deferrable | None = None,
        include: list[str] | tuple[str, ...] | None = None,
        violation_error_code: str | None = None,
        violation_error_message: _StrOrPromise | None = None,
    ) -> None: ...
    @override
    def check(self, model: type[Model], connection: BaseDatabaseWrapper) -> list[CheckMessage]: ...
    @override
    def validate(
        self, model: type[Model], instance: Model, exclude: Iterable[str] | None = None, using: str = "default"
    ) -> None: ...

__all__ = ["ExclusionConstraint"]



from typing import Any

from mypy.checkmember import analyze_member_access
from mypy.errorcodes import ATTR_DEFINED
from mypy.nodes import CallExpr, MemberExpr
from mypy.plugin import AttributeContext
from mypy.types import AnyType, Instance, TypeOfAny
from mypy.types import Type as MypyType
from mypy.version import __version__ as mypy_version

from mypy_django_plugin.lib import helpers

mypy_version_info = tuple(map(int, mypy_version.partition("+")[0].split(".")))


def resolve_str_promise_attribute(ctx: AttributeContext) -> MypyType:
    if isinstance(ctx.context, MemberExpr):
        method_name = ctx.context.name
    elif isinstance(ctx.context, CallExpr) and isinstance(ctx.context.callee, MemberExpr):
        method_name = ctx.context.callee.name
    else:
        ctx.api.fail(f'Cannot resolve the attribute of "{ctx.type}"', ctx.context, code=ATTR_DEFINED)
        return AnyType(TypeOfAny.from_error)

    str_info = helpers.lookup_fully_qualified_typeinfo(helpers.get_typechecker_api(ctx), "builtins.str")
    assert str_info is not None
    str_type = Instance(str_info, [])

    # TODO: [mypy 1.16+] Remove this workaround for passing `msg` to `analyze_member_access()`.
    extra: dict[str, Any] = {}
    if mypy_version_info < (1, 16):
        extra["msg"] = ctx.api.msg

    return analyze_member_access(
        method_name,
        str_type,
        ctx.context,
        is_lvalue=False,
        is_super=False,
        # operators are already handled with magic methods defined in the stubs for _StrPromise
        is_operator=False,
        original_type=ctx.type,
        chk=helpers.get_typechecker_api(ctx),
        **extra,
    )



from argparse import Action, ArgumentParser, Namespace
from collections.abc import Sequence
from typing import Any

from django.core.management.base import BaseCommand
from typing_extensions import override

class LayerOptionAction(Action):
    @override
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        value: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None: ...

class ListOptionAction(Action):
    @override
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        value: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None: ...

class Command(BaseCommand):
    @override
    def add_arguments(self, parser: ArgumentParser) -> None: ...
    @override
    def handle(self, *args: Any, **options: Any) -> str: ...



from collections.abc import Callable, Collection, Iterable, Mapping, Sequence, Sized
from typing import Any, TypeVar, overload

from django.template.base import FilterExpression, Parser
from django.template.context import Context
from django.utils.safestring import SafeString

from .base import Node, NodeList, Template

class InvalidTemplateLibrary(Exception): ...

_C = TypeVar("_C", bound=Callable[..., Any])

class Library:
    filters: dict[str, Callable[..., Any]]
    tags: dict[str, Callable[..., Any]]
    def __init__(self) -> None: ...
    @overload
    def tag(self, name: _C) -> _C: ...
    @overload
    def tag(self, name: str, compile_function: _C) -> _C: ...
    @overload
    def tag(self, name: str | None = None, compile_function: None = None) -> Callable[[_C], _C]: ...
    def tag_function(self, func: _C) -> _C: ...
    @overload
    def filter(self, name: _C, filter_func: None = None, **flags: Any) -> _C: ...
    @overload
    def filter(self, name: str | None, filter_func: _C, **flags: Any) -> _C: ...
    @overload
    def filter(self, name: str | None = None, filter_func: None = None, **flags: Any) -> Callable[[_C], _C]: ...
    def filter_function(self, func: _C, **flags: Any) -> _C: ...
    @overload
    def simple_tag(self, func: _C, takes_context: bool | None = None, name: str | None = None) -> _C: ...
    @overload
    def simple_tag(
        self, func: None = None, takes_context: bool | None = None, name: str | None = None
    ) -> Callable[[_C], _C]: ...
    @overload
    def simple_block_tag(
        self,
        func: _C,
        takes_context: bool | None = None,
        name: str | None = None,
        end_name: str | None = None,
    ) -> _C: ...
    @overload
    def simple_block_tag(
        self,
        func: None = None,
        takes_context: bool | None = None,
        name: str | None = None,
        end_name: str | None = None,
    ) -> Callable[[_C], _C]: ...
    @overload
    def inclusion_tag(
        self,
        filename: Template | str,
        func: _C,
        takes_context: bool | None = None,
        name: str | None = None,
    ) -> _C: ...
    @overload
    def inclusion_tag(
        self,
        filename: Template | str,
        func: None = None,
        takes_context: bool | None = None,
        name: str | None = None,
    ) -> Callable[[_C], _C]: ...

class TagHelperNode(Node):
    func: Any
    takes_context: Any
    args: Any
    kwargs: Any
    def __init__(
        self,
        func: Callable[..., Any],
        takes_context: bool | None,
        args: list[FilterExpression],
        kwargs: dict[str, FilterExpression],
    ) -> None: ...
    def get_resolved_arguments(self, context: Context) -> tuple[list[int], dict[str, SafeString | int]]: ...

class SimpleNode(TagHelperNode):
    args: list[FilterExpression]
    func: Callable[..., Any]
    kwargs: dict[str, FilterExpression]
    takes_context: bool | None
    target_var: str | None
    def __init__(
        self,
        func: Callable[..., Any],
        takes_context: bool | None,
        args: list[FilterExpression],
        kwargs: dict[str, FilterExpression],
        target_var: str | None,
    ) -> None: ...

class SimpleBlockNode(SimpleNode):
    nodelist: NodeList

    def __init__(
        self,
        nodelist: NodeList,
        *args: Any,
        **kwargs: Any,
    ) -> None: ...

class InclusionNode(TagHelperNode):
    args: list[FilterExpression]
    func: Callable[..., Any]
    kwargs: dict[str, FilterExpression]
    takes_context: bool | None
    filename: Template | str
    def __init__(
        self,
        func: Callable[..., Any],
        takes_context: bool | None,
        args: list[FilterExpression],
        kwargs: dict[str, FilterExpression],
        filename: Template | str | None,
    ) -> None: ...

def parse_bits(
    parser: Parser,
    bits: Iterable[str],
    params: Sequence[str],
    varargs: str | None,
    varkw: str | None,
    defaults: Sized | None,
    kwonly: Collection[str],
    kwonly_defaults: Mapping[str, int] | None,
    takes_context: bool | None,
    name: str,
) -> tuple[list[FilterExpression], dict[str, FilterExpression]]: ...
def import_library(name: str) -> Library: ...



from collections.abc import Iterable, Iterator, Sequence
from typing import Any, Literal, overload

from django.core.checks.messages import CheckMessage
from django.core.files.storage import FileSystemStorage, Storage
from typing_extensions import override

searched_locations: Any

class BaseFinder:
    def check(self, **kwargs: Any) -> list[CheckMessage]: ...
    @overload
    def find(self, path: str, find_all: Literal[False] = False) -> str | None: ...
    @overload
    def find(self, path: str, find_all: Literal[True]) -> list[str]: ...
    def list(self, ignore_patterns: Iterable[str] | None) -> Iterable[Any]: ...

class FileSystemFinder(BaseFinder):
    locations: list[tuple[str, str]]
    storages: dict[str, Any]
    def __init__(self, app_names: Sequence[str] | None = None, *args: Any, **kwargs: Any) -> None: ...
    @overload
    @override
    def find(self, path: str, find_all: Literal[False] = False) -> str | None: ...
    @overload
    @override
    def find(self, path: str, find_all: Literal[True]) -> list[str]: ...
    def find_location(self, root: str, path: str, prefix: str | None = None) -> str | None: ...
    @override
    def list(self, ignore_patterns: Iterable[str] | None) -> Iterable[Any]: ...

class AppDirectoriesFinder(BaseFinder):
    storage_class: type[FileSystemStorage]
    source_dir: str
    apps: list[str]
    storages: dict[str, FileSystemStorage]
    def __init__(self, app_names: Iterable[str] | None = None, *args: Any, **kwargs: Any) -> None: ...
    @overload
    @override
    def find(self, path: str, find_all: Literal[False] = False) -> str | None: ...
    @overload
    @override
    def find(self, path: str, find_all: Literal[True]) -> list[str]: ...
    def find_in_app(self, app: str, path: str) -> str | None: ...
    @override
    def list(self, ignore_patterns: Iterable[str] | None) -> Iterable[Any]: ...

class BaseStorageFinder(BaseFinder):
    storage: Storage | None
    def __init__(self, storage: Storage | None = None, *args: Any, **kwargs: Any) -> None: ...
    @overload
    @override
    def find(self, path: str, find_all: Literal[False] = False) -> str | None: ...
    @overload
    @override
    def find(self, path: str, find_all: Literal[True]) -> list[str]: ...
    @override
    def list(self, ignore_patterns: Iterable[str] | None) -> Iterable[Any]: ...

class DefaultStorageFinder(BaseStorageFinder):
    storage: Storage
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...

@overload
def find(path: str, find_all: Literal[False] = False) -> str | None: ...
@overload
def find(path: str, find_all: Literal[True]) -> list[str]: ...
def get_finders() -> Iterator[BaseFinder]: ...
def get_finder(import_path: str) -> BaseFinder: ...



from collections.abc import Iterator
from typing import Any

from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models.expressions import Expression, Func
from django.db.models.fields import Field
from django.db.models.lookups import Exact, GreaterThan, GreaterThanOrEqual, In, IsNull, LessThan, LessThanOrEqual
from django.db.models.sql.compiler import SQLCompiler, _AsSqlType
from typing_extensions import override

class Tuple(Func):
    function: str
    output_field: Field
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[Expression]: ...
    @override
    def as_sqlite(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...  # type: ignore[override]

class TupleLookupMixin:
    allows_composite_expressions: bool
    def get_prep_lookup(self) -> Any: ...
    def check_rhs_is_tuple_or_list(self) -> None: ...
    def check_rhs_length_equals_lhs_length(self) -> None: ...
    def check_rhs_is_supported_expression(self) -> None: ...
    def get_lhs_str(self) -> str: ...
    def get_prep_lhs(self) -> Any: ...
    def process_lhs(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper, lhs: Any = None) -> _AsSqlType: ...
    def process_rhs(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...
    def get_fallback_sql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...
    def as_sql(self, compiler: SQLCompiler, connection: BaseDatabaseWrapper) -> _AsSqlType: ...

class TupleExact(TupleLookupMixin, Exact): ...
class TupleIsNull(TupleLookupMixin, IsNull): ...
class TupleGreaterThan(TupleLookupMixin, GreaterThan): ...
class TupleGreaterThanOrEqual(TupleLookupMixin, GreaterThanOrEqual): ...
class TupleLessThan(TupleLookupMixin, LessThan): ...
class TupleLessThanOrEqual(TupleLookupMixin, LessThanOrEqual): ...

class TupleIn(TupleLookupMixin, In):
    def check_rhs_is_collection_of_tuples_or_lists(self) -> None: ...
    def check_rhs_elements_length_equals_lhs_length(self) -> None: ...
    def check_rhs_is_query(self) -> None: ...

tuple_lookups: dict[str, type[TupleLookupMixin]]
